# ----------------------------------------------------------------------
# ctokens.py
#
# Token specifications for symbols in ANSI C and C++.  This file is
# meant to be used as a library in other tokenizers.
# ----------------------------------------------------------------------

# flake8:noqa
# isort:skip_file
# type: ignore

# Reserved words

tokens = [
    # Literals (identifier, integer constant, float constant, string constant, char const)
    "ID",
    "TYPEID",
    "INTEGER",
    "FLOAT",
    "STRING",
    "CHARACTER",
    # Operators (+,-,*,/,%,|,&,~,^,<<,>>, ||, &&, !, <, <=, >, >=, ==, !=)
    "PLUS",
    "MINUS",
    "TIMES",
    "DIVIDE",
    "MODULO",
    "OR",
    "AND",
    "NOT",
    "XOR",
    "LSHIFT",
    "RSHIFT",
    "LOR",
    "LAND",
    "LNOT",
    "LT",
    "LE",
    "GT",
    "GE",
    "EQ",
    "NE",
    # Assignment (=, *=, /=, %=, +=, -=, <<=, >>=, &=, ^=, |=)
    "EQUALS",
    "TIMESEQUAL",
    "DIVEQUAL",
    "MODEQUAL",
    "PLUSEQUAL",
    "MINUSEQUAL",
    "LSHIFTEQUAL",
    "RSHIFTEQUAL",
    "ANDEQUAL",
    "XOREQUAL",
    "OREQUAL",
    # Increment/decrement (++,--)
    "INCREMENT",
    "DECREMENT",
    # Structure dereference (->)
    "ARROW",
    # Ternary operator (?)
    "TERNARY",
    # Delimeters ( ) [ ] { } , . ; :
    "LPAREN",
    "RPAREN",
    "LBRACKET",
    "RBRACKET",
    "LBRACE",
    "RBRACE",
    "COMMA",
    "PERIOD",
    "SEMI",
    "COLON",
    # Ellipsis (...)
    "ELLIPSIS",
]

# Operators
t_PLUS = r"\+"
t_MINUS = r"-"
t_TIMES = r"\*"
t_DIVIDE = r"/"
t_MODULO = r"%"
t_OR = r"\|"
t_AND = r"&"
t_NOT = r"~"
t_XOR = r"\^"
t_LSHIFT = r"<<"
t_RSHIFT = r">>"
t_LOR = r"\|\|"
t_LAND = r"&&"
t_LNOT = r"!"
t_LT = r"<"
t_GT = r">"
t_LE = r"<="
t_GE = r">="
t_EQ = r"=="
t_NE = r"!="

# Assignment operators

t_EQUALS = r"="
t_TIMESEQUAL = r"\*="
t_DIVEQUAL = r"/="
t_MODEQUAL = r"%="
t_PLUSEQUAL = r"\+="
t_MINUSEQUAL = r"-="
t_LSHIFTEQUAL = r"<<="
t_RSHIFTEQUAL = r">>="
t_ANDEQUAL = r"&="
t_OREQUAL = r"\|="
t_XOREQUAL = r"\^="

# Increment/decrement
t_INCREMENT = r"\+\+"
t_DECREMENT = r"--"

# ->
t_ARROW = r"->"

# ?
t_TERNARY = r"\?"

# Delimeters
t_LPAREN = r"\("
t_RPAREN = r"\)"
t_LBRACKET = r"\["
t_RBRACKET = r"\]"
t_LBRACE = r"\{"
t_RBRACE = r"\}"
t_COMMA = r","
t_PERIOD = r"\."
t_SEMI = r";"
t_COLON = r":"
t_ELLIPSIS = r"\.\.\."

# Identifiers
t_ID = r"[A-Za-z_][A-Za-z0-9_]*"

# Integer literal
t_INTEGER = r"\d+([uU]|[lL]|[uU][lL]|[lL][uU])?"

# Floating literal
t_FLOAT = r"((\d+)(\.\d+)(e(\+|-)?(\d+))? | (\d+)e(\+|-)?(\d+))([lL]|[fF])?"

# String literal
t_STRING = r"\"([^\\\n]|(\\.))*?\""

# Character constant 'c' or L'c'
t_CHARACTER = r"(L)?\'([^\\\n]|(\\.))*?\'"


# Comment (C-Style)
def t_COMMENT(t):
    r"/\*(.|\n)*?\*/"
    t.lexer.lineno += t.value.count("\n")
    return t


# Comment (C++-Style)
def t_CPPCOMMENT(t):
    r"//.*\n"
    t.lexer.lineno += 1
    return t



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2019-05-13
# @Filename: actor.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import logging
import pathlib
import warnings

from typing import Any, Callable, Dict, List, Optional, Tuple, TypeVar, Union, cast

import click

import clu
from clu.parsers.click import CluCommand

from ..actor import CustomTransportType
from ..base import BaseActor, MessageCode, Reply
from ..command import Command, TimedCommandList, parse_legacy_command
from ..parsers import ClickParser
from ..protocol import TCPStreamServer
from ..tools import log_reply
from .tron import TronConnection
from .types.messages import Reply as OpsReply


__all__ = ["LegacyActor", "BaseLegacyActor"]


T = TypeVar("T", bound="BaseLegacyActor")
PathLike = Union[str, pathlib.Path]


@click.command(cls=CluCommand)
async def tron_reconnect(*args):
    """Reconnects to tron/hub."""

    command = args[0]

    if command.actor.tron is None:
        return command.fail("Tron instance not set.")

    command.actor.tron.stop()
    await asyncio.sleep(0.5)

    await command.actor.tron.start()

    return command.finish()


class BaseLegacyActor(BaseActor):
    """An actor that provides compatibility with the SDSS opscore protocol.

    The TCP servers need to be started by awaiting the coroutine `.start`.
    Note that `.start` does not block so you will need to use asyncio's
    `.run_forever` or a similar system ::

        >>> loop = asyncio.get_event_loop()
        >>> my_actor = await LegacyActor('my_actor', '127.0.0.1', 9999, loop=loop)
        >>> my_actor.start()
        >>> loop.run_forever()

    Parameters
    ----------
    name
        The name of the actor.
    host
        The host where the TCP server will run.
    port
        The port of the TCP server.
    tron_host
        The host on which Tron is running.
    tron_port
        The port on which Tron is running.
    models
        A list of strings with the actors whose models will be tracked.
    version
        The version of the actor.
    loop
        The event loop. If `None`, the current event loop will be used.
    log_dir
        The directory where to store the logs. Defaults to
        ``$HOME/logs/<name>`` where ``<name>`` is the name of the actor.
    log
        A `~logging.Logger` instance to be used for logging instead of creating
        a new one.
    schema
        The path to the datamodel schema for the actor, in JSON Schema format.
        If the schema is provided all replies will be validated against it.
        An invalid reply will fail and not be emitted. The schema can also be
        set when subclassing by setting the class ``schema`` attribute.
    store
        Whether to store the output keywords in a `.KeywordStore`. `False`
        (the default), disables the feature. `True` will store a record
        of all the output keywords. A list of keyword names to store can
        also be passed.
    config
        A dictionary of configuration parameters that will be accessible to the actor.
    """

    def __init__(
        self,
        name: str,
        host: str,
        port: int,
        tron_host: Optional[str] = None,
        tron_port: Optional[int] = None,
        models: List[str] = [],
        version: Optional[str] = None,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        log_dir: Optional[PathLike] = None,
        log: Optional[logging.Logger] = None,
        verbose: bool = False,
        schema: Optional[PathLike] = None,
        store: bool | list[str] = False,
        additional_properties: bool = False,
        config: Dict[str, Any] = {},
    ):
        super().__init__(
            name,
            version=version,
            log_dir=log_dir,
            log=log,
            verbose=verbose,
            schema=schema,
            additional_properties=additional_properties,
            config=config,
            store=store,
        )

        #: Mapping of user_id to transport
        self.transports = dict()

        self.host = host
        self.port = port

        # TCPStreamServer: The server to talk to this actor.
        self._server = TCPStreamServer(
            host,
            port,
            connection_callback=self.new_user,
            data_received_callback=self.new_command,
        )

        if tron_host and tron_port:
            #: TronConnection: The client connection to Tron.
            self.tron = TronConnection(
                f"{self.name}.{self.name}",
                host=tron_host,
                port=tron_port,
                models=models,
                log=self.log,
            )
        else:
            self.tron = None

        #: dict: Actor models.
        self.models = self.tron.models if self.tron else {}

        self.timed_commands = TimedCommandList(self)

    def __repr__(self):
        return (
            f"<{str(self)} (name={self.name!r}, host={self.host!r}, port={self.port})>"
        )

    async def start(self: T, get_keys: bool = True, start_nubs: bool = True) -> T:
        """Starts the server and the Tron client connection.

        Parameters
        ----------
        get_keys
            Whether to issue ``keys getFor`` commands against the hub to retrieve
            the current values for the model keys.
        start_nubs
            If `True`, and a `.TronConnection` has been created, sends a
            ``hub startNubs <name>`` where ``<name>`` is the name of the
            actor to attempt to automatically connect it to the hub.

        """

        self.set_loop_exception_handler()

        await self._server.start()
        self.log.info(f"running TCP server on {self.host}:{self.port}")

        # Start tron connection
        try:
            if self.tron:
                await self.tron.start(get_keys=get_keys)
                self.log.info(
                    f"started tron connection at {self.tron.host}:{self.tron.port}"
                )
                if start_nubs:
                    self.log.debug("Asking Tron to connect back.")
                    await self.send_command("hub", f"startNubs {self.name}")
            else:
                warnings.warn(
                    "starting LegacyActor without Tron connection.",
                    clu.CluWarning,
                )
        except (ConnectionRefusedError, OSError) as ee:
            warnings.warn(
                f"connection to tron was refused: {ee}. "
                "Some functionality will be limited.",
                clu.CluWarning,
            )

        self.timed_commands.start()

        return self

    async def stop(self):
        """Stops the client connection and running tasks."""

        if self._server.is_serving():
            self._server.stop()

        await self.timed_commands.stop()

        if self.tron:
            self.tron.stop()

    async def run_forever(self):
        """Runs the actor forever, keeping the loop alive."""

        await self._server.serve_forever()

    def new_user(self, transport: CustomTransportType):
        """Assigns userID to new client connection."""

        if transport.is_closing():
            if hasattr(transport, "user_id"):
                self.log.debug(f"user {transport.user_id} disconnected.")
                return self.transports.pop(transport.user_id)

        curr_ids: set[int] = set(self.transports.keys())
        user_id = 1 if len(curr_ids) == 0 else max(curr_ids) + 1

        transport.user_id = user_id

        self.transports[user_id] = transport

        # report user information and additional info
        self.show_new_user_info(user_id)

        return

    def new_command(self, transport: CustomTransportType, command_str: bytes):
        """Handles a new command received by the actor."""

        user_id = getattr(transport, "user_id", 0)
        command_str_s = command_str.decode().strip()

        if not command_str_s:
            return
        try:
            commander_id, command_id, command_body = parse_legacy_command(command_str_s)

            command = Command(
                command_string=command_body,
                commander_id=commander_id or user_id,
                command_id=command_id,
                consumer_id=self.name,
                actor=self,
                transport=transport,
            )
        except clu.CommandError:
            self.write(
                "f",
                {"error": f"Could not parse command {command_str_s!r}"},
            )
            return

        self.log.info(
            f"New command received: {command_body   !r} "
            f"(commander_id={commander_id!r}, command_id={command_id!r})"
        )

        return self.parse_command(command)

    @staticmethod
    def format_user_output(
        user_id: int,
        command_id: int,
        message_code: MessageCode,
        msg_str: Optional[str] = None,
    ) -> str:
        """Formats a string to send to users."""

        msg_str = "" if msg_str is None else " " + msg_str

        return f"{user_id} {command_id:d} {message_code.value:s}{msg_str:s}"

    def show_new_user_info(self, user_id: int):
        """Shows information for new users. Called when a new user connects."""

        self.show_user_info(user_id)
        self.show_version(user_id=user_id)

        transport = self.transports[user_id]
        peername = transport.get_extra_info("peername")[0]
        self.log.debug(f"user {user_id} connected from {peername!r}.")

    def show_user_info(self, user_id: int):
        """Shows user information including your user_id."""

        num_users = len(self.transports)
        if num_users == 0:
            return

        msg = {"yourUserID": user_id, "num_users": num_users}

        self.write("i", msg, user_id=user_id)
        self.show_user_list()

    def show_user_list(self):
        """Shows a list of connected users. Broadcast to all users."""

        user_id_list = sorted(self.transports.keys())
        for user_id in user_id_list:
            transport = self.transports[user_id]
            peername = transport.get_extra_info("peername")[0]
            msg = {"UserInfo": [user_id, peername]}
            self.write("i", msg)

    @staticmethod
    def get_user_command_id(
        command: Optional[Command] = None,
        user_id: int = 0,
        command_id: int = 0,
    ) -> Tuple[int, int]:
        """Returns commander_id, command_id based on user-supplied information.

        Parameters
        ----------
        command
            User command; used as a default for ``user_id`` and
            ``command_id``.
        user_id
            If `None` then use ``command.user_id``.
        command_id
            If `None` then use ``command.command_id``.

        Returns
        -------
        user_id, command_id
            The commander ID and the command ID, parsed from the inputs. If
            they cannot be determined, returns zeros.
        """

        cid = cast(int, command.command_id) if command else 0
        user_id = user_id or cid
        command_id = command_id or cid
        return (user_id, command_id)

    def show_version(self, user_id: int = 0):
        """Shows actor version."""

        msg = {"version": repr(self.version)}

        self.write("i", msg, user_id=user_id)

    def send_command(
        self,
        target: str,
        command_string: str,
        *args,
        commander: Optional[str] = None,
        command_id: Optional[int] = None,
        command: Optional[Command] = None,
        callback: Optional[Callable[[OpsReply], None]] = None,
        time_limit: Optional[float] = None,
    ):
        """Sends a command through the hub.

        Parameters
        ----------
        target
            The actor to command.
        command_string
            The command to send.
        args
            Arguments to concatenate to the command string.
        commander
            The commander string to send to Tron. If not provided, a valid
            string is built using the name of the actor and the target.
        command_id
            The command id. If `None`, a sequentially increasing value will
            be used. You should not specify a ``command_id`` unless you really
            know what you're doing.
        callback
            A callback to invoke with each reply received from the actor.
        time_limit
            A delay after which the command is marked as timed out and done.

        Examples
        --------
        These two are equivalent ::

            >>> actor.send_command('my_actor', 'do_something --now')
            >>> actor.send_command('my_actor', 'do_something', '--now')

        """

        if command and isinstance(command.commander_id, str):
            commander = command.commander_id
        else:
            commander = None

        if self.tron and self.tron.connected():
            command = self.tron.send_command(
                target,
                command_string,
                *args,
                commander=commander,
                mid=command_id,
                callback=callback,
                time_limit=time_limit,
            )
            return command

        else:
            raise clu.CluError("cannot connect to tron.")

    def write(
        self,
        message_code: MessageCode | str = "i",
        message: Optional[Dict[str, Any]] = None,
        command: Optional[Command] = None,
        user_id: int = 0,
        command_id: int = 0,
        concatenate: bool = True,
        broadcast: bool = False,
        validate: bool = True,
        write_to_log: bool = True,
        **kwargs,
    ):
        """Writes a message to user(s).

        Parameters
        ----------
        message_code
            The message code (e.g., ``'i'`` or ``':'``).
        message
            The keywords to be output. Must be a dictionary of pairs
            ``{keyword: value}``. If ``value`` is a list it will be converted
            into a comma-separated string. To prevent unexpected casting,
            it is recommended for ``value`` to always be a string.
        command
            User command; used as a default for ``user_id`` and
            ``command_id``. If the command is done, it is ignored.
        user_id
            The user (transport) to which to write. `None` defaults to 0.
        command_id
            If `None` then use ``command.command_id``.
        concatenate
            Concatenates all the keywords to be output in a single
            reply with the keyword-values joined with semicolons. Otherwise
            each keyword will be output as a different message.
        broadcast
            Whether to broadcast the reply. Equivalent to ``user_id=0``.
        validate
            Validate the reply against the actor model. This is ignored if the actor
            was not started with knowledge of its own schema.
        write_to_log
            Whether to write the reply to the log. Defaults to yes but
            it may be useful to prevent large repetitive replies cluttering
            the log.
        kwargs
            Keyword arguments that will be added to the message. If a keyword
            is both in ``message`` and in ``kwargs``, the value in ``kwargs``
            supersedes ``message``.
        """

        message_code = MessageCode(message_code)

        reply = BaseActor.write(
            self,
            message_code=message_code,
            message=message,
            command=command,
            broadcast=broadcast,
            validate=validate,
            emit=False,
            expand_exceptions=False,
            **kwargs,
        )

        if kwargs.get("silent", False) is False:
            self._write_internal(
                reply,
                user_id=user_id,
                command_id=command_id,
                concatenate=concatenate,
                write_to_log=write_to_log,
            )

    def _write_internal(
        self,
        reply: Reply,
        user_id=0,
        command_id=0,
        concatenate=True,
        write_to_log: bool = True,
    ):
        """Writes reply to users.

        Parameters
        ----------
        reply
            The reply object to output to users.
        user_id
            The user to which we are replying.
        command_id
            The command emitting this message.
        write_to_log
            Whether to write the reply to the log. Defaults to yes but
            it may be useful to prevent large repetitive replies cluttering
            the log.

        """

        command = cast(Command, reply.command)
        message = reply.message

        # For a reply, the commander ID is the user assigned to the transport
        # that issues this command.
        transport = command.transport if command else None
        user_id = (transport.user_id if transport else None) or user_id

        if reply.broadcast:
            user_id = 0
            command_id = 0

        user_id, command_id = self.get_user_command_id(
            command=command,
            user_id=user_id,
            command_id=command_id,
        )

        lines = []
        for keyword in message:
            try:
                value = clu.format_value(message[keyword])
            except BaseException as err:
                raise TypeError(f"Cannot format keyword {keyword!r} " + str(err))

            if value.strip() == "":
                lines.append(f"{keyword}")
            else:
                lines.append(f"{keyword}={value}")

        if concatenate:
            lines = ["; ".join(lines)]

        for line in lines:
            full_msg_str = self.format_user_output(
                user_id,
                command_id,
                reply.message_code,
                line,
            )
            msg = (full_msg_str + "\n").encode()

            if user_id is None or user_id == 0 or transport is None:
                for transport in self.transports.values():
                    transport.write(msg)
            else:
                global_transport = self.transports.get(transport.user_id, None)
                if global_transport is not None and global_transport == transport:
                    transport.write(msg)

            if self.log and write_to_log:
                log_reply(self.log, reply.message_code, full_msg_str)


class LegacyActor(ClickParser, BaseLegacyActor):
    """A legacy actor that uses the `.ClickParser`."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.tron:
            self.parser.add_command(tron_reconnect)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2019-05-20
# @Filename: base.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import abc
import asyncio
import enum
import inspect
import logging
import pathlib
import time
from datetime import datetime, timezone

from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Optional,
    Type,
    TypeVar,
    Union,
    cast,
)

import jsonschema.exceptions
import yaml
from pydantic import BaseModel

from sdsstools import get_logger, read_yaml_file
from sdsstools.logger import SDSSLogger

from .model import Model
from .store import KeywordStore
from .tools import REPLY


if TYPE_CHECKING:
    from clu.command import BaseCommand, Command
    from clu.legacy.types.messages import Keywords


__all__ = ["BaseClient", "BaseActor", "Reply", "MessageCode"]


SchemaType = Union[Type[BaseModel], Dict[str, Any], pathlib.Path, str]
T = TypeVar("T", bound="BaseClient")


class MessageCode(enum.Enum):
    """Flags for message codes."""

    STARTED = ">"
    DONE = ":"
    FAILED = "f"
    ERROR = "e"
    CRITICAL = "!"
    INFO = "i"
    WARNING = "w"
    DEBUG = "d"

    def __eq__(self, __o: object) -> bool:
        if isinstance(__o, MessageCode):
            return super().__eq__(__o)

        return self.value == __o


class BaseClient(metaclass=abc.ABCMeta):
    """A base client that can be used for listening or for an actor.

    This class defines a new client. Clients differ from actors in that
    they do not receive commands or issue replies, but do send commands to
    other actors and listen to the keyword-value flow. All actors are also
    clients and any actor should subclass from `.BaseClient`.

    Normally a new instance of a client or actor is created by passing a
    configuration file path to `.from_config` which defines how the
    client must be started.

    Parameters
    ----------
    name
        The name of the client.
    version
        The version of the client.
    log_dir
        The directory where to store the file logs.
    log
        A `~logging.Logger` instance to be used for logging instead of creating
        a new one.
    verbose
        Whether to log to stdout. Can be an integer logging level.
    validate
        Whether to actor should validate its own messages against its model (if it
        has one). This is a global parameter that can be overridden when calling
        `~.BaseClient.write`.
    config
        A dictionary of configuration parameters that will be accessible to the client.

    """

    name: str

    command_models: dict[str, dict] = {}

    def __init__(
        self,
        name: str,
        version: Optional[str] = None,
        log_dir: Optional[Union[pathlib.Path, str]] = None,
        log: Optional[SDSSLogger] = None,
        verbose: Union[bool, int] = False,
        validate: bool = True,
        config: dict = {},
    ):
        self.name = name
        assert self.name, "name cannot be empty."

        self.log: SDSSLogger
        self.setup_logger(log, log_dir, verbose=verbose)

        self.version = version or "?"

        # Internally store the original configuration used to start the client.
        self.config: Dict[str, Any] = {}

        self.validate = validate
        self.config = config

    def __repr__(self):
        return f"<{str(self)} (name={self.name!r})>"

    def __str__(self):
        return self.__class__.__name__

    @abc.abstractmethod
    async def start(self: T) -> T:
        """Runs the client."""

        pass

    async def stop(self):
        """Shuts down all the remaining tasks."""

        self.log.info("cancelling all pending tasks and shutting down.")

        tasks = [
            task for task in asyncio.all_tasks() if task is not asyncio.current_task()
        ]
        list(map(lambda task: task.cancel(), tasks))

        await asyncio.gather(*tasks, return_exceptions=True)

        asyncio.get_running_loop().stop()

    @staticmethod
    def _parse_config(
        input: Union[Dict[str, Any], pathlib.Path, str],
        loader=yaml.FullLoader,
    ) -> Dict[str, Any]:
        if not isinstance(input, dict):
            input = pathlib.Path(input)
            assert input.exists(), "configuration path does not exist."
            config = read_yaml_file(str(input), loader=loader)
        else:
            config = input

        return cast("Dict[str, Any]", config)

    @classmethod
    def from_config(
        cls,
        config: Union[Dict[str, Any], pathlib.Path, str],
        *args,
        loader=yaml.FullLoader,
        **kwargs,
    ):
        """Parses a configuration file.

        Parameters
        ----------
        config
            A configuration dictionary or the path to a YAML configuration
            file. If the file contains a section called ``'actor'`` or
            ``'client'``, that section will be used instead of the whole
            file.
        """

        orig_config_dict = cls._parse_config(config, loader=loader)

        config_dict = orig_config_dict.copy()

        if "actor" in config_dict:
            config_dict = config_dict["actor"]
        elif "client" in config_dict:
            config_dict = config_dict["client"]

        config_dict.update(kwargs)

        # Decide what to do with the rest of the keyword arguments:
        args_inspect = inspect.getfullargspec(cls)

        if args_inspect.varkw is not None:
            # If there is a catch-all kw variable, send everything and let the
            # subclass handle it.
            config_dict.update(kwargs)
        else:
            # Check the kw arguments in the subclass and pass only
            # values from config_dict that match them.
            class_kwargs = args_inspect.args
            class_kwargs.remove("self")

            # Remove keys that are not in the signature.
            config_dict = {
                key: value for key, value in config_dict.items() if key in class_kwargs
            }

        # We also pass *args in case the actor has been subclassed
        # and the subclass' __init__ accepts different arguments.
        new_client = cls(*args, config=orig_config_dict, **config_dict)

        return new_client

    def setup_logger(
        self,
        log: Any,
        log_dir: Optional[Union[pathlib.Path, str]],
        verbose: Union[bool, int] = False,
    ):
        """Starts the file logger."""

        if not log:
            log = get_logger("clu:" + self.name)
        else:
            assert isinstance(log, SDSSLogger), "Logger must be sdsstools.SDSSLogger"

        log.setLevel(REPLY)

        if log_dir:
            log_dir = pathlib.Path(log_dir).expanduser()

            log.start_file_logger(
                str(log_dir / f"{self.name}.log"),
                rotating=True,
                rollover=True,
            )

            if log.fh:  # In case starting the file logger fails.
                if log.fh.formatter:
                    log.fh.formatter.converter = time.gmtime
                log.fh.setLevel(REPLY)

        log.sh.setLevel(logging.WARNING)
        if verbose is True:
            log.sh.setLevel(logging.DEBUG)
        elif verbose is not False and isinstance(verbose, int):
            log.sh.setLevel(verbose)

        self.log = log
        self.log.debug(f"{self.name}: logging system initiated.")

        return log

    def set_loop_exception_handler(self):
        """Sets the lopp exception handler to be handled by the logger."""

        if self.log:
            # Set the loop exception handler to be handled by the logger.
            loop = asyncio.get_running_loop()
            loop.set_exception_handler(self.log.asyncio_exception_handler)

    async def send_command(self, actor: str, *args, **kwargs) -> Command:
        """Sends a command to an actor and returns a `.Command` instance."""

        raise NotImplementedError(
            "Sending commands is not implemented for this client."
        )

    def proxy(self, actor: str) -> ProxyClient:
        """Creates a proxy for an actor.

        Returns a `.ProxyClient` that simplifies running a command multiple times.
        For example ::

            await client.send_command("focus_state", "moverelative -1000 UM")

        can be replaced with ::

            focus_stage = client.proxy("focus_stage")
            await focus_stage.send_command("moverelative", "-1000", "UM")

        """

        return ProxyClient(self, actor)


class ProxyClient:
    """A proxy representing an actor.

    Parameters
    ----------
    client
        The client used to command the actor.
    actor
        The actor to command.

    """

    def __init__(self, client: BaseClient, actor: str):
        self.client = client
        self.actor = actor

    def send_command(self, *args):
        """Sends a command to the actor.

        Returns the result of calling the client ``send_command()`` method
        with the actor and concatenated arguments as parameters. Note that
        in some cases the client ``send_command()`` method may be a coroutine
        function, in which case the returned coroutine needs to be awaited.

        Parameters
        ----------
        args
            Arguments to pass to the actor. They will be concatenated using spaces.

        """

        command = " ".join(map(str, args))

        return self.client.send_command(self.actor, command)


class BaseActor(BaseClient):
    """An actor based on `asyncio`.

    This class expands `.BaseClient` with a parsing system for new commands
    and placeholders for methods for handling new commands and writing replies,
    which should be overridden by the specific actors.

    Parameters
    ----------
    schema
        The schema for the actor replies, as a JSONschema dictionary or path
        to a JSON file. If `None`, defaults to the internal basic schema.
    store
        Whether to store the output keywords in a `.KeywordStore`. `False`
        (the default), disables the feature. `True` will store a record
        of all the output keywords. A list of keyword names to store can
        also be passed.
    additional_properties
        Whether to allow additional properties in the schema, other than the
        ones defined by the schema. This parameter only is used if
        ``schema=None`` or if ``additionalProperties`` is not defined in
        the schema.

    """

    model: Union[Model, None] = None
    _message_processor: Callable[[dict], dict] | None = None

    def __init__(
        self,
        *args,
        schema: SchemaType | None = None,
        store: bool | list[str] = False,
        additional_properties: bool = False,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self.load_schema(schema, additional_properties=additional_properties)

        if store is False:
            self.store = None
        else:
            self.store = KeywordStore(self, filter=None if store is True else store)

    def load_schema(
        self,
        schema: Union[SchemaType, None],
        is_file=None,
        additional_properties=False,
    ) -> Union[Model, None]:
        """Loads and validates the actor schema."""

        if schema is None:
            schema = {
                "$schema": "http://json-schema.org/draft-07/schema#",
                "type": "object",
                "properties": {},
                "additionalProperties": additional_properties,
            }
            is_file = False

        self.model = Model(
            self.name,
            schema,
            is_file=is_file,
            additional_properties=additional_properties,
        )

        return self.model

    def set_message_processor(self, processor: Callable[[dict], dict] | None):
        """Sets the message processor.

        Parameters
        ----------
        processor
            A function that receives the messsages to output to the users, as a
            dictionary, and reformats them, returning a new dictionary. If `None`,
            no processing is done.

        """

        self._message_processor = processor

    @abc.abstractmethod
    def new_command(self):
        """Handles a new command.

        Must be overridden by the subclass and call `.parse_command`
        with a `.Command` object.
        """
        pass

    @abc.abstractmethod
    def parse_command(self, command: BaseCommand):
        """Parses and executes a `.Command`. Must be overridden."""
        pass

    @abc.abstractmethod
    def _write_internal(self, reply: Reply, write_to_log: bool = True):
        """Internally handle the reply and output it to the users.

        Must handle converting the general `.Reply` to the specific format of the
        actor transport. Must also handle logging the reply.

        Parameters
        ----------
        reply
            The reply object to output to users.
        write_to_log
            Whether to write the reply to the log. Defaults to yes but
            it may be useful to prevent large repetitive replies cluttering
            the log.

        """

        pass

    def write(
        self,
        message_code: MessageCode | str = MessageCode.INFO,
        message: Optional[Dict[str, Any] | str] = None,
        command: Optional[BaseCommand] = None,
        broadcast: bool = False,
        validate: bool | None = None,
        expand_exceptions: bool = True,
        traceback_frame: int = 0,
        silent: bool = False,
        internal: bool = False,
        emit: bool = True,
        write_to_log: bool = True,
        **kwargs,
    ) -> Reply:
        """Writes a message to user(s).

        The reply to the users will be formatted by the actor class into message
        specific to the communication channel. Additional keywords passed to this
        method will be used to complete the message (as long as they don't overlap
        with named parameters). For example ::

            actor.write('i', message={'text': 'Hi!', 'value1': 1})

        and ::

            actor.write('i', text='Hi!', value1=1)

        are equivalent and ::

            actor.write('i', message={'text': 'Hi!'}, value1=1)

        is equivalent to ::

            actor.write('i', message={'text': 'Hi!', 'value1': 1})

        This method generates a `.Reply` object that is passed to the
        ``_write_internal`` method in the class, which processes it and outputs the
        message to the users using the appropriate transport. If ``command`` is passed,
        the reply is added to ``command.replies``.

        Parameters
        ----------
        message_code
            The message code (e.g., ``'i'`` or ``':'``).
        message
            The keywords to be output. Must be a dictionary of pairs
            ``{keyword: value}``.
        command
            The command to which we are replying. If not set, it is assumed
            that this is a broadcast.
        broadcast
            Whether to broadcast the message to all the users or only to the
            commander.
        validate
            Validate the reply against the actor schema. This is ignored if the actor
            was not started with knowledge of its own schema. If `None`, defaults to
            the actor global behaviour.
        expand_exceptions
            If the value of one of the keywords is an exception object and
            `expand_exception=True`, the exception information and traceback will
            be expanded. Otherwise the exception will just be stringified.
        traceback_frame
            The frame from the traceback to use if ``expand_exception=True``. By
            default reports the file and line of the last traceback frame.
        silent
            When `True` does not output the message to the users. This can be used to
            issue internal commands that update the internal model but that don't
            clutter the output.
        internal
            Marks the `.Reply` instance as internal. Internal replies are expected to
            be emitted to the users, but with a header indicating that the message
            is not for broader consumption. This can be useful when one wants to
            reply with verbose information that can be skipped from CLI or logs.
        emit
            Whether to call the actor internal write method. Should be `True` but
            it's sometimes useful to call `.write` with ``emit=False`` when
            one is overriding the method and wants to control when to call the
            internal method. In that case, a `.Reply` object is returned but nothing
            is output to the users.
        write_to_log
            Whether to write the reply to the log. Defaults to yes but
            it may be useful to prevent large repetitive replies cluttering
            the log.
        kwargs
            Keyword arguments that will used to update the message.
        """

        message_code = MessageCode(message_code)

        if isinstance(message, str):
            message = {"text": message}
        else:
            message = message or {}
            if not isinstance(message, dict):
                raise TypeError("message must be a string or a dictionary.")

        message.update(kwargs)

        if self._message_processor:
            message = self._message_processor(message)

        for key, value in message.items():
            if isinstance(value, Exception):
                if expand_exceptions is True:
                    filename: str | None = None
                    lineno: int | None = None
                    if value.__traceback__ is not None:
                        tb = value.__traceback__
                        for _ in range(traceback_frame):
                            t_next = tb.tb_next
                            if t_next is None:
                                break
                            tb = t_next

                        filename = tb.tb_frame.f_code.co_filename if tb else None
                        lineno = tb.tb_lineno if tb else None

                    message[key] = {
                        "module": value.__class__.__module__,
                        "type": value.__class__.__name__,
                        "message": str(value),
                        "filename": filename,
                        "lineno": lineno,
                    }
                else:
                    message[key] = str(value)

        reply = Reply(
            message_code,
            message,
            command=command,
            broadcast=broadcast,
            internal=internal,
        )

        do_validate = validate if validate is not None else self.validate
        if do_validate and self.model is not None:
            reply.use_validation = True
            result, err = self.model.validate(message, update_model=True)
            if result is False:
                if isinstance(err, jsonschema.exceptions.ValidationError):
                    message = {
                        "error": f"Failed validating the reply: message {message} "
                        "does not match the schema."
                    }
                else:
                    message = {"error": f"Failed validating the reply: {err}"}

                reply.message_code = MessageCode.ERROR
                reply.message = message
                reply.validated = False
            else:
                reply.validated = True

        if command:
            command.replies.append(reply)

        if emit and silent is False:
            if asyncio.iscoroutinefunction(self._write_internal):
                asyncio.create_task(
                    self._write_internal(
                        reply,
                        write_to_log=write_to_log,
                    )
                )
            else:
                self._write_internal(reply, write_to_log=write_to_log)

        if self.store is not None:
            self.store.add_reply(reply)

        return reply

    def invoke_mock_command(self, command_str, command_id=0) -> Command:
        """Send a new command to an actor for testing.

        Requires calling `.setup_test_actor`."""

        raise NotImplementedError("setup_test_actor() has not been called.")


class Reply:
    """A reply from a command or actor to be sent to the users."""

    def __init__(
        self,
        message_code: MessageCode | str,
        message: Dict[str, Any],
        command: Optional[BaseCommand] = None,
        broadcast: bool = False,
        use_validation: bool = False,
        validated: bool = False,
        internal: bool = False,
        keywords: Optional[Keywords] = None,
    ):
        self.date = datetime.now(timezone.utc)
        self.message_code = MessageCode(message_code)
        self.message = message
        self.command = command
        self.broadcast = broadcast
        self.use_validation = use_validation
        self.validated = validated
        self.keywords = keywords
        self.internal = internal

    @property
    def body(self):
        """Alias to ``message``."""

        return self.message



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2022-09-01
# @Filename: store.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from datetime import datetime

from typing import TYPE_CHECKING, Any


if TYPE_CHECKING:
    from .base import BaseActor, Reply

__all__ = ["KeywordStore", "KeywordOutput"]


class KeywordStore(defaultdict):
    """Stores the keywords output by an actor.

    Parameters
    ----------
    actor
        The actor to which this store is attached to.
    filter
        A list of keyword names to filter. If provided, only those keywords
        will be tracked.

    """

    def __init__(self, actor: BaseActor, filter: list[str] | None = None):
        self.actor = actor
        self.name = self.actor.name

        self.filter = filter

        defaultdict.__init__(self, list)

    def add_reply(self, reply: Reply):
        """Processes a reply and adds new entries to the store.

        Parameters
        ----------
        reply
            The `.Reply` object containing the keywords output in a message
            from the actor.

        """

        for keyword, value in reply.message.items():
            if self.filter is not None and keyword not in self.filter:
                continue

            key_out = KeywordOutput(keyword, reply.message_code, datetime.now(), value)

            if keyword in self:
                self[keyword].append(key_out)
            else:
                self[keyword] = [key_out]

    def head(self, keyword: str, n: int = 1):
        """Returns the first N output values of a keyword.

        Parameters
        ----------
        keyword
            The name of the keyword to search for.
        n
            Return the first ``n`` times the keyword was output.

        """

        return self[keyword][:n]

    def tail(self, keyword: str, n: int = 1):
        """Returns the last N output values of a keyword.

        Parameters
        ----------
        keyword
            The name of the keyword to search for.
        n
            Return the last ``n`` times the keyword was output.

        """

        return self[keyword][-n:]


@dataclass
class KeywordOutput:
    """Records a single output of a keyword.

    Parameters
    ----------
    name
        The name of the keyword.
    message_code
        The message code with which the keyword was output.
    date
        A `.datetime` object with the date-time at which the keyword was
        output this time.
    value
        The value of the keyword when it was output.

    """

    name: str
    message_code: Any
    date: datetime
    value: Any



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2018-01-19
# @Filename: command.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import random
from contextlib import suppress

from typing import Any, Callable, Dict, List, Optional, TypeVar, Union

import aio_pika as apika
import aiormq

from .exceptions import CluError


__all__ = [
    "TCPProtocol",
    "PeriodicTCPServer",
    "TCPStreamServer",
    "TCPStreamPeriodicServer",
    "TCPStreamClient",
    "open_connection",
    "TopicListener",
    "ReconnectingTCPClientProtocol",
]

T = TypeVar("T")
ConnectionCallbackType = Callable[[Any], Any]
DataReceivedCallbackType = Callable[[Any, bytes], Any]


class TCPProtocol(asyncio.Protocol):
    """A TCP server/client based on asyncio protocols.

    This is a high-level implementation of the client and server asyncio
    protocols. See `asyncio protocol
    <https://docs.python.org/3/library/asyncio-protocol.html>`__ for details.

    Parameters
    ----------
    loop
        The event loop. The current event loop is used by default.
    connection_callback
        Callback to call when a new client connects.
    data_received_callback
        Callback to call when a new data is received.
    max_connections
        How many clients the server accepts. If `None`, unlimited connections
        are allowed.
    """

    def __init__(
        self,
        loop: asyncio.AbstractEventLoop | None = None,
        connection_callback: Optional[ConnectionCallbackType] = None,
        data_received_callback: Optional[Callable[[str], Any]] = None,
        max_connections: Optional[int] = None,
    ):
        self.connection_callback = connection_callback
        self.data_received_callback = data_received_callback

        self.transports = []
        self.max_connections = max_connections

        self.loop = loop or asyncio.get_event_loop()

    @classmethod
    async def create_server(cls, host: str, port: int, **kwargs):
        """Returns a `~asyncio.Server` connection."""

        loop = kwargs.get("loop", asyncio.get_event_loop())

        new_tcp = cls(**kwargs)

        server = await loop.create_server(lambda: new_tcp, host, port)

        await server.start_serving()

        return server

    @classmethod
    async def create_client(cls, host: str, port: int, **kwargs):
        """Returns a `~asyncio.Transport` and `~asyncio.Protocol`."""

        if "connection_callback" in kwargs:
            raise KeyError("connection_callback not allowed when creating a client.")

        loop = kwargs.get("loop", asyncio.get_event_loop())

        new_tcp = cls.__new__(cls, **kwargs)
        transport, protocol = await loop.create_connection(lambda: new_tcp, host, port)

        return transport, protocol

    def connection_made(self, transport: asyncio.Transport):
        """Receives a connection and calls the connection callback."""

        if self.max_connections is None or (
            len(self.transports) < self.max_connections
        ):
            self.transports.append(transport)
        else:
            transport.write(b"Maximum number of connections reached.")
            transport.close()

        if self.connection_callback:
            self.connection_callback(transport)

    def data_received(self, data: bytes):
        """Decodes the received data."""

        if self.data_received_callback:
            self.data_received_callback(data.decode())

    def connection_lost(self, exc):
        """Called when connection is lost."""

        pass


class PeriodicTCPServer(TCPProtocol):
    """A TCP server that runs a callback periodically.

    Parameters
    ----------
    period_callback
        Callback to run every iteration.
    sleep_time
        The delay between two calls to ``periodic_callback``.
    kwargs
        Parameters to pass to `TCPProtocol`

    """

    def __init__(
        self,
        periodic_callback: Optional[ConnectionCallbackType] = None,
        sleep_time: float = 1,
        **kwargs,
    ):
        self._periodic_callback = periodic_callback
        self.sleep_time = sleep_time

        self.periodic_task = None

        super().__init__(**kwargs)

    @classmethod
    async def create_client(cls, *args, **kwargs):
        raise NotImplementedError(
            "create_client is not implemented for PeriodicTCPServer."
        )

    @classmethod
    async def create_server(cls, host: str, port: int, *args, **kwargs):
        """Returns a `~asyncio.Server` connection."""

        loop = kwargs.get("loop", asyncio.get_event_loop())

        new_tcp = cls(*args, **kwargs)

        server = await loop.create_server(lambda: new_tcp, host, port)

        await server.start_serving()

        new_tcp.periodic_task = asyncio.create_task(new_tcp._emit_periodic())

        return server

    @property
    def periodic_callback(self) -> Optional[ConnectionCallbackType]:
        """Returns the periodic callback."""

        return self._periodic_callback

    @periodic_callback.setter
    def periodic_callback(self, func: Callable[[asyncio.Transport], Any]):
        """Sets the periodic callback."""

        self._periodic_callback = func

    async def _emit_periodic(self):
        while True:
            if self.periodic_callback is not None:
                for transport in self.transports:
                    if asyncio.iscoroutinefunction(self.periodic_callback):
                        await self.periodic_callback(transport)
                    else:
                        self.periodic_callback(transport)

            await asyncio.sleep(self.sleep_time)


class TCPStreamServer(object):
    """A TCP server based on asyncio streams.

    This is a high-level implementation of the asyncio server using
    streams. See `asyncio streams
    <https://docs.python.org/3/library/asyncio-stream.html>`__ for details.

    Parameters
    ----------
    host
        The server host.
    port
        The server port.
    connection_callback
        Callback to call when a new client connects or disconnects.
    data_received_callback
        Callback to call when a new data is received.
    loop
        The event loop. The current event loop is used by default.
    max_connections
        How many clients the server accepts. If `None`, unlimited connections
        are allowed.
    """

    def __init__(
        self,
        host: str,
        port: int,
        connection_callback: Optional[ConnectionCallbackType] = None,
        data_received_callback: Optional[DataReceivedCallbackType] = None,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        max_connections: Optional[int] = None,
    ):
        self.host = host
        self.port = port

        self.transports = {}
        self.loop = loop or asyncio.get_event_loop()

        self.max_connections = max_connections

        self.connection_callback = connection_callback
        self.data_received_callback = data_received_callback

        # The `asyncio.Server`. Created when `.start_server` is run.
        self._server = None

    async def start(self) -> asyncio.AbstractServer:
        """Starts the server and returns a `~asyncio.Server` connection."""

        self._server = await asyncio.start_server(
            self.connection_made,
            self.host,
            self.port,
        )

        return self._server

    def stop(self):
        """Stops the server."""

        assert self._server
        self._server.close()

    def serve_forever(self):
        """Exposes ``TCPStreamServer.server.serve_forever``."""

        assert self._server
        return self._server.serve_forever()

    def is_serving(self) -> bool:
        assert self._server
        return self._server.is_serving()

    async def _do_callback(self, cb, *args, **kwargs):
        """Calls a function or coroutine callback."""

        if asyncio.iscoroutinefunction(cb):
            return await asyncio.create_task(cb(*args, **kwargs))
        else:
            return cb(*args, **kwargs)

    async def connection_made(
        self,
        reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
    ):
        """Called when a new client connects to the server.

        Stores the writer protocol in ``transports``, calls the connection
        callback, if any, and starts a loop to read any incoming data.
        """

        if self.max_connections and len(self.transports) == self.max_connections:
            writer.write("Max number of connections reached.\n".encode())
            await writer.drain()
            return

        self.transports[writer.transport] = writer

        if self.connection_callback:
            await self._do_callback(self.connection_callback, writer.transport)

        while True:
            try:
                data = await reader.readuntil()
            except (asyncio.IncompleteReadError, ConnectionResetError):
                break

            if data == b"" or reader.at_eof():
                break

            if self.data_received_callback:
                await self._do_callback(
                    self.data_received_callback,
                    writer.transport,
                    data,
                )

        self.transports.pop(writer.transport)
        writer.close()

        if self.connection_callback:
            await self._do_callback(self.connection_callback, writer.transport)

        with suppress(ConnectionResetError):
            await writer.wait_closed()


class TCPStreamClient:
    """An object containing a writer and reader stream to a TCP server."""

    def __init__(self, host: str, port: int):
        self.host = host
        self.port = port

        self.reader = None
        self.writer = None

    async def open_connection(self):
        """Creates the connection."""

        self.reader, self.writer = await asyncio.open_connection(self.host, self.port)

    def close(self):
        """Closes the stream."""

        if self.writer:
            self.writer.close()
        else:
            raise RuntimeError("connection cannot be closed because it is not open.")


async def open_connection(host: str, port: int) -> TCPStreamClient:
    """Returns a TCP stream connection with a writer and reader.

    This function is equivalent to doing ::

        >>> client = TCPStreamClient('127.0.0.1', 5555)
        >>> await client.open_connection()

    Instead just do ::

        >>> client = await TCPStreamClient('127.0.0.1', 5555)
        >>> client.writer.write('Hi!\\n'.encode())

    Parameters
    ----------
    host : str
        The host of the TCP server.
    port : int
        The port of the TCP server.

    Returns
    -------
    client : `.TCPStreamClient`
        A container for the stream reader and writer.
    """

    client = TCPStreamClient(host, port)
    await client.open_connection()

    return client


class TCPStreamPeriodicServer(TCPStreamServer):
    """A TCP server that calls a function periodically.

    Parameters
    ----------
    host
        The server host.
    port
        The server port.
    period_callback
        Callback to run every iteration. It is called for each transport
        that is connected to the server and receives the transport object.
    sleep_time
        The delay between two calls to ``periodic_callback``.
    kwargs
        Parameters to pass to `TCPStreamServer`

    """

    def __init__(
        self,
        host: str,
        port: int,
        periodic_callback: Optional[Callable[[asyncio.Transport], Any]] = None,
        sleep_time: float = 1,
        **kwargs,
    ):
        self._periodic_callback = periodic_callback
        self.sleep_time = sleep_time

        self.periodic_task = None

        super().__init__(host, port, **kwargs)

    async def start(self) -> asyncio.AbstractServer:
        """Starts the server and returns a `~asyncio.Server` connection."""

        self._server = await super().start()

        self.periodic_task = asyncio.create_task(self._emit_periodic())

        return self._server

    def stop(self):
        if self.periodic_task:
            self.periodic_task.cancel()

        super().stop()

    @property
    def periodic_callback(self):
        """Returns the periodic callback."""

        return self._periodic_callback

    @periodic_callback.setter
    def periodic_callback(self, func: Callable[[asyncio.Transport], Any]):
        """Sets the periodic callback."""

        self._periodic_callback = func

    async def _emit_periodic(self):
        while True:
            if self._server and self.periodic_callback:
                for transport in self.transports:
                    await self._do_callback(self.periodic_callback, transport)

            await asyncio.sleep(self.sleep_time)


class TopicListener(object):
    """A class to declare and listen to AMQP queues with topic conditions.

    Parameters
    ----------
    url
        RFC3986 formatted broker address. When used, the other keyword
        arguments are ignored.
    user
        The user to connect to the RabbitMQ broker.
    password
        The password for the user.
    host
        The host where the RabbitMQ message broker runs.
    virtualhost
         Virtualhost parameter. ``'/'`` by default.
    port
        The port on which the RabbitMQ message broker is running.
    ssl
        Whether to use TLS/SSL connection.
    """

    def __init__(
        self,
        url: str | None = None,
        user: str = "guest",
        password: str = "guest",
        host: str = "localhost",
        virtualhost: str = "/",
        port: int = 5672,
        ssl: bool = False,
    ):
        self.url = url
        self.user = user
        self.password = password
        self.host = host
        self.port = port
        self.virtualhost = virtualhost
        self.ssl = ssl

        self.connection: apika.abc.AbstractConnection | None = None
        self.channel: apika.abc.AbstractChannel
        self.exchange: apika.abc.AbstractExchange
        self.queues: List[apika.abc.AbstractQueue] = []

        self._consumer_tag: Dict[apika.abc.AbstractQueue, apika.queue.ConsumerTag] = {}

    async def connect(
        self,
        exchange_name: str,
        exchange_type: apika.ExchangeType = apika.ExchangeType.TOPIC,
        on_return_raises=True,
    ) -> TopicListener:
        """Initialise the connection.

        Parameters
        ----------
        exchange_name
            The name of the exchange to create.
        exchange_type
            The type of exchange to create.
        """

        try:
            if self.url:
                self.connection = await apika.connect(self.url)
            else:
                self.connection = await apika.connect(
                    login=self.user,
                    host=self.host,
                    port=self.port,
                    password=self.password,
                    virtualhost=self.virtualhost,
                    ssl=self.ssl,
                )
        except ConnectionError as err:
            raise ConnectionError(f"Failed conneting to the AMQP server: {err}.")

        self.channel = await self.connection.channel(on_return_raises=on_return_raises)
        await self.channel.set_qos(prefetch_count=1)

        self.exchange = await self.channel.declare_exchange(
            exchange_name,
            type=exchange_type,
            auto_delete=True,
        )

        return self

    async def add_queue(
        self,
        queue_name: str,
        callback: Optional[Callable[[apika.abc.AbstractIncomingMessage], Any]] = None,
        bindings: Union[str, List[str]] = "*",
    ) -> apika.abc.AbstractQueue:
        """Adds a queue with bindings.

        Parameters
        ----------
        queue_name
            The name of the queue to create.
        callback
            A callable that will be called when a new message is received in
            the queue. Can be a coroutine.
        bindings
            The list of bindings for the queue. Can be a list of string or a
            single string in which the bindings are comma-separated.
        """

        if isinstance(bindings, str):
            bindings = bindings.split(",")
        elif isinstance(bindings, (list, tuple)):
            bindings = list(bindings)
        else:
            raise TypeError(f"invalid type for bindings {bindings!r}.")

        try:
            queue = await self.channel.declare_queue(queue_name, exclusive=True)
        except aiormq.exceptions.ChannelLockedResource:
            raise CluError(
                f"cannot create queue {queue_name}. "
                "This may indicate that another instance of the "
                "same actor is running."
            )

        for binding in bindings:
            await queue.bind(self.exchange, routing_key=binding)

        if callback:
            self._consumer_tag[queue] = await queue.consume(callback)

        self.queues.append(queue)

        return queue

    async def stop(self):
        """Cancels queues and closes the connection."""

        for queue in self.queues:
            consumer_tag = self._consumer_tag.get(queue, None)
            if hasattr(queue, "consumer_tag") and consumer_tag is not None:
                await queue.cancel(consumer_tag)

        if self.connection:
            await self.connection.close()


class ReconnectingTCPClientProtocol(asyncio.Protocol):
    """A reconnecting client modelled after Twisted ``ReconnectingClientFactory``.

    Taken from https://bit.ly/3yn6MWa.
    """

    max_delay = 3600
    initial_delay = 1.0
    factor = 2.7182818284590451
    jitter = 0.119626565582
    max_retries = None

    def __init__(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs
        self._retries = 0
        self._delay = self.initial_delay
        self._continue_trying = True
        self._call_handle = None
        self._connector = None
        self.connected = False

    def connection_lost(self, exc):
        self.connected = False
        if self._continue_trying:
            self.retry()

    def connection_failed(self, exc):
        self.connected = False
        if self._continue_trying:
            self.retry()

    def retry(self):
        if not self._continue_trying:
            return

        self._retries += 1
        if self.max_retries is not None and (self._retries > self.max_retries):
            return

        self._delay = min(self._delay * self.factor, self.max_delay)
        if self.jitter:
            self._delay = random.normalvariate(self._delay, self._delay * self.jitter)
        self._call_handle = asyncio.get_event_loop().call_later(
            self._delay, self.connect
        )

    def connect(self):
        if self._connector is None:
            self._connector = asyncio.create_task(self._connect())

    async def _connect(self):
        try:
            await asyncio.get_event_loop().create_connection(
                lambda: self,
                *self._args,
                **self._kwargs,
            )
            self.connected = True
        except Exception as exc:
            asyncio.get_event_loop().call_soon(self.connection_failed, exc)
        finally:
            self._connector = None

    def stop_trying(self):
        if self._call_handle:
            self._call_handle.cancel()
            self._call_handle = None
        self._continue_trying = False
        if self._connector is not None:
            self._connector.cancel()
            self._connector = None
        self.connected = False



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2019-04-22
# @Filename: __init__.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

# flake8: noqa

import logging

from sdsstools import get_package_version

from .actor import *
from .base import *
from .client import *
from .command import *
from .device import *
from .exceptions import *
from .legacy import LegacyActor
from .parsers import *
from .tools import (
    REPLY,
    ActorHandler,
    CommandStatus,
    as_complete_failer,
    escape,
    format_value,
)


# Add REPLY level to logging
logging.addLevelName(REPLY, "REPLY")
logging.Logger.REPLY = lambda self, message, *args, **kws: self._log(  # type: ignore
    REPLY, message, *args, **kws
)


NAME = "sdss-clu"
__version__ = get_package_version(__file__, "sdss-clu", pep_440=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-07-30
# @Filename: client.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import json
import logging
import pathlib
import uuid
import warnings
from copy import deepcopy

from typing import TYPE_CHECKING, Any, Awaitable, Callable, Dict, List, Optional, Union

import aio_pika as apika

from sdsstools.logger import SDSSLogger

from clu.exceptions import CluWarning

from .base import BaseClient, Reply
from .command import Command
from .model import ModelSet
from .protocol import TopicListener
from .tools import CommandStatus


if TYPE_CHECKING:
    from aio_pika.abc import HeadersType


__all__ = ["AMQPClient", "AMQPReply"]


PathLike = Union[str, pathlib.Path]
ReplyCallbackType = Callable[["AMQPReply"], Union[None, Awaitable[None]]]


class AMQPReply(object):
    """Wrapper for an `~aio_pika.IncomingMessage` that expands and decodes it.

    Parameters
    ----------
    message
        The message that contains the reply.
    log
        A message logger.

    Attributes
    ----------
    is_valid
        Whether the message is valid and correctly parsed.
    body
        The body of the message, as a JSON dictionary.
    info
        The info dictionary.
    headers
        The headers of the message, decoded if they are bytes.
    message_code
        The message code.
    internal
        Whether this reply was marked internal.
    sender
        The name of the actor that sends the reply.
    command_id
        The command ID.
    """

    def __init__(
        self,
        message: apika.abc.AbstractIncomingMessage,
        log: Optional[logging.Logger] = None,
    ):
        self.command_id: str | None = None
        self.message_code: str = ""
        self.body = {}

        self.message = message
        self._log = log

        self.is_valid = True
        self.info = message.info()

        self.headers = dict(self.info.get("headers", {}))

        self.message_code = str(self.headers.get("message_code", ""))
        if self.message_code == "":
            self.is_valid = False
            if self._log:
                self._log.warning(f"received message without message_code: {message}")
            return

        self.internal = bool(self.headers.get("internal", False))

        self.sender = self.headers.get("sender", None)
        if self.sender is None and self._log:
            self._log.warning(f"received message without sender: {message}")

        self.command_id = message.correlation_id

        command_id_header = self.headers.get("command_id", None)
        if command_id_header and command_id_header != self.command_id:
            if self._log:
                self._log.error(
                    f"mismatch between message "
                    f"correlation_id={self.command_id} "
                    f"and header command_id={command_id_header} "
                    f"in message {message}"
                )
            self.is_valid = False
            return

        self.body = json.loads(self.message.body)


class AMQPClient(BaseClient):
    """Defines a new client based on the AMQP standard.

    To start a new client first instantiate the class and then run `.start` as
    a coroutine. Note that `.start` does not block so you will need to use
    asyncio's ``run_forever`` or a similar system ::

        >>> loop = asyncio.get_event_loop()
        >>> client = await AMQPClient('my_client', host='localhost').start()
        >>> loop.run_forever()

    Parameters
    ----------
    name
        The name of the client.
    url
        RFC3986 formatted broker address. When used, the other connection
        keyword arguments are ignored.
    user
        The user to connect to the AMQP broker. Defaults to ``guest``.
    password
        The password for the user. Defaults to ``guest``.
    host
        The host where the AMQP message broker runs. Defaults to ``localhost``.
    virtualhost
         Virtualhost parameter. ``'/'`` by default.
    port
        The port on which the AMQP broker is running. Defaults to 5672.
    ssl
        Whether to use TLS/SSL connection.
    version
        The version of the client.
    loop
        The event loop. If `None`, the current event loop will be used.
    log_dir
        The directory where to store the logs. Defaults to
        ``$HOME/logs/<name>`` where ``<name>`` is the name of the actor.
    log
        A `~logging.Logger` instance to be used for logging instead of
        creating a new one.
    parser
        A click command parser that is a subclass of `~clu.parser.CluGroup`.
        If `None`, the active parser will be used.
    models
        A list of actor models whose schemas will be monitored.
    """

    __EXCHANGE_NAME__ = "sdss_exchange"

    def __init__(
        self,
        name: str | None = None,
        url: Optional[str] = None,
        user: str = "guest",
        password: str = "guest",
        host: str = "localhost",
        port: int = 5672,
        virtualhost: str = "/",
        ssl: bool = False,
        version: Optional[str] = None,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        log_dir: Optional[PathLike] = None,
        log: Optional[SDSSLogger] = None,
        models: List[str] = [],
        **kwargs,
    ):
        if name is None:
            name = "amqp-client-" + str(uuid.uuid4()).split("-")[-1]

        super().__init__(
            name,
            version=version,
            log_dir=log_dir,
            log=log,
            **kwargs,
        )

        self.replies_queue = None

        # Creates the connection to the AMQP broker
        self.connection = TopicListener(
            url=url,
            user=user,
            password=password,
            host=host,
            port=port,
            ssl=ssl,
            virtualhost=virtualhost,
        )

        #: dict: External commands currently running.
        self.running_commands: Dict[str, Command] = {}

        self.models = ModelSet(self, actors=models, raise_exception=False)

        self._callbacks: list[ReplyCallbackType] = []

    def __repr__(self):
        if self.connection.connection is None:
            url = "disconnected"
        else:
            assert isinstance(self.connection.connection, apika.Connection)
            url = str(self.connection.connection.url)

        return f"<{str(self)} (name={self.name!r}, {url})>"

    async def start(self, exchange_name: str = __EXCHANGE_NAME__):
        """Starts the connection to the AMQP broker."""

        self.set_loop_exception_handler()

        # Starts the connection and creates the exchange
        await self.connection.connect(exchange_name)

        # Binds the replies queue.
        self.replies_queue = await self.connection.add_queue(
            f"{self.name}_replies",
            callback=self.handle_reply,
            bindings=["reply.#"],
        )

        if self.connection.connection:
            assert isinstance(self.connection.connection, apika.Connection)
            url = self.connection.connection.url
        else:
            url = "???"

        self.log.debug(f"Replies queue {self.replies_queue.name!r} bound to {url!s}")

        # Initialises the models.
        await self.models.load_schemas()

        return self

    async def stop(self):
        """Cancels queues and closes the connection."""

        if self.connection.connection and not self.connection.connection.is_closed:
            await self.connection.stop()

    def is_connected(self):
        """Is the client connected to the exchange?"""

        return self.connection.connection and not self.connection.connection.is_closed

    async def run_forever(self):
        """Runs the event loop forever."""

        assert self.connection.connection

        while not self.connection.connection.is_closed:
            await asyncio.sleep(1)

    async def __aenter__(self):
        """Starts the client inside a context manager."""

        try:
            if not self.is_connected():
                await self.start()
        except Exception:
            await self.stop()

    async def __aexit__(self, *_):
        """Exits the context manager."""

        await self.stop()

    async def handle_reply(
        self,
        message: apika.abc.AbstractIncomingMessage,
    ) -> AMQPReply:
        """Handles a reply received from the exchange.

        Creates a new instance of `.AMQPReply` from the ``message``. If the
        reply is valid it updates any running command.

        Parameters
        ----------
        message
            The message received.

        Returns
        -------
        reply
            The `.AMQPReply` object created from the message.
        """

        reply = AMQPReply(message, log=self.log)
        await reply.message.ack()

        if not reply.is_valid:
            self.log.error("Invalid message received.")
            return reply

        # Update the models
        if self.models and reply.sender in self.models:
            self.models[reply.sender].update_model(reply.body)

        # If the command is running we check if the message code indicates
        # the command is done and, if so, sets the result in the Future.
        # Also, add the reply to the command list of replies.
        if reply.command_id and reply.command_id in self.running_commands:
            command = self.running_commands[reply.command_id]
            command.replies.append(
                Reply(
                    message=reply.body,
                    message_code=reply.message_code,
                    command=command,
                    internal=reply.internal,
                    validated=True,
                )
            )
            if command._reply_callback is not None:
                command._reply_callback(reply)
            status = CommandStatus.code_to_status(reply.message_code)
            if command.status != status:
                command.set_status(status)
            if status.is_done:
                if not command.done():
                    command.set_result(command)
                    del self.running_commands[reply.command_id]

        # Handle reply callbacks.
        for cb in self._callbacks:
            if asyncio.iscoroutinefunction(cb):
                asyncio.create_task(cb(reply))
            else:
                cb(reply)

        return reply

    async def send_command(
        self,
        consumer: str,
        command_string: str,
        *args,
        command_id: str | None = None,
        callback: Optional[Callable[[AMQPReply], None]] = None,
        internal: bool = False,
        command: Optional[Command] = None,
        time_limit: Optional[float] = None,
        await_command: bool = True,
    ):
        """Commands another actor over its RCP queue.

        Parameters
        ----------
        consumer
            The actor we are commanding.
        command_string
            The command string that will be parsed by the remote actor.
        args
            Arguments to concatenate to the command string.
        command_id
            The command ID associated with this command. If empty, an unique
            identifier will be attached.
        callback
            A callback to invoke with each reply received from the actor.
        internal
            Whether to mark the command as internal, in which case replies will
            also be considered internal.
        command
            The `.Command` that initiated the new command. Only relevant for
            actors.
        time_limit
            A delay after which the command is marked as timed out and done.
        await_command
            If `True`, awaits the command until it finishes.

        Examples
        --------
        These two are equivalent ::

            >>> client.send_command('my_actor', 'do_something --now')
            >>> client.send_command('my_actor', 'do_something', '--now')

        """

        if command and command.command_id:
            command_id = str(command.command_id)
        else:
            command_id = command_id or str(uuid.uuid4())

        if len(args) > 0:
            command_string += " " + " ".join(map(str, args))

        if command and isinstance(command.commander_id, str):
            commander_id = command.commander_id + f".{consumer}"
        else:
            commander_id = f"{self.name}.{consumer}"

        internal = command.internal if command else internal

        # Creates and registers a command.
        command = Command(
            command_string=command_string,
            command_id=command_id,
            commander_id=commander_id,
            consumer_id=consumer,
            internal=internal,
            actor=None,
            reply_callback=callback,
            time_limit=time_limit,
        )

        self.running_commands[command_id] = command

        headers = {
            "command_id": command_id,
            "commander_id": commander_id,
            "internal": internal,
        }

        message_body = {"command_string": command_string}

        await self._publish_message(
            consumer,
            headers=headers,
            body=message_body,
            correlation_id=command_id,
        )

        if await_command:
            await command

        return command

    async def _publish_message(
        self,
        consumer: str,
        headers: HeadersType = {},
        body: dict[str, Any] = {},
        correlation_id: str | None = None,
    ):
        """Publishes a message to an exchange."""

        if not hasattr(self.connection, "exchange"):
            warnings.warn(
                f"Exchange is not ready to output message: {body}",
                CluWarning,
            )
            return

        assert self.replies_queue

        # The routing key has the topic command and the name of the commanded actor.
        routing_key = f"command.{consumer}"

        try:
            await self.connection.exchange.publish(
                apika.Message(
                    json.dumps(body).encode(),
                    content_type="text/json",
                    headers=headers,
                    correlation_id=correlation_id,
                    reply_to=self.replies_queue.name,
                ),
                routing_key=routing_key,
            )
        except (apika.exceptions.DeliveryError, apika.exceptions.PublishError):
            # The consumer (actor) did not reply. This usually means that the actor
            # is not connected. We fake a reply from that actor saying so. That will
            # be received by handle_reply which will fail the current command.
            error_msg = dict(error=f"Failed routing message to consumer {consumer!r}.")
            headers.update({"message_code": "f", "sender": consumer})
            await self.connection.exchange.publish(
                apika.Message(
                    json.dumps(error_msg).encode(),
                    content_type="text/json",
                    headers=headers,
                    correlation_id=correlation_id,
                ),
                routing_key=f"reply.{self.name}",
            )

    async def send_task(
        self,
        consumer: str,
        task_name: str,
        payload: dict[str, Any] = {},
        **kwargs,
    ):
        """

        Parameters
        ----------
        consumer
            The actor we are commanding.
        task_name
            The task to execute in the remote actor
        payload
            A serialisable dictionary with the payload to pass to the task.
        kwargs
            Additional arguments used to update the payload dictionary.

        """

        assert self.replies_queue

        payload = deepcopy(payload)
        payload.update(kwargs)

        headers = {"commander_id": f"{self.name}.{consumer}", "task": True}

        message_body = {"task": task_name}
        message_body.update(payload)

        correlation_id = str(uuid.uuid4())

        await self._publish_message(
            consumer,
            headers=headers,
            body=message_body,
            correlation_id=correlation_id,
        )

    def add_reply_callback(self, callback_func: ReplyCallbackType):
        """Adds a callback that is called when a new reply is received."""

        if callback_func not in self._callbacks:
            self._callbacks.append(callback_func)

    def remove_reply_callback(self, callback_func: ReplyCallbackType):
        """Removes a reply callback."""

        if callback_func not in self._callbacks:
            raise ValueError("Callback not registered.")

        self._callbacks.remove(callback_func)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2018-01-17
# @Filename: command.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import logging
import re
import sys
import time
import warnings
from contextlib import suppress

from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
    cast,
)

import click

import clu
import clu.base
from clu.exceptions import CluWarning, CommandError
from clu.tools import CommandStatus, StatusMixIn


__all__ = [
    "BaseCommand",
    "Command",
    "parse_legacy_command",
    "TimedCommand",
    "TimedCommandList",
    "FakeCommand",
]


Actor_co = TypeVar("Actor_co", bound="clu.base.BaseActor")
Future_co = TypeVar("Future_co", bound="BaseCommand")
Reply_co = TypeVar("Reply_co", bound="clu.base.Reply")

if sys.version_info >= (3, 9, 0):
    Future = asyncio.Future
else:

    class Future(asyncio.Future, Generic[Future_co]):
        pass


class ReplyList(List[Reply_co]):
    """A list of replies to a command."""

    def get(self, keyword: str):
        """Return the value of the reply that matches the key."""

        for reply in self:
            if keyword in reply.message:
                return reply.message[keyword]

        raise KeyError(f"Keyword {keyword} not found.")


class BaseCommand(
    Future[Future_co],
    StatusMixIn[CommandStatus],
    Generic[Actor_co, Future_co],
):
    """Base class for commands of all types (user and device).

    A `BaseCommand` instance is a `~asyncio.Future` whose result gets set
    when the status is done (either successfully or not).

    Parameters
    ----------
    commander_id
        The ID of the commander issuing this command. Can be a string or an
        integer. Normally the former is used for new-style actor and the latter
        for legacy actors.
    command_id
        The ID associated to this command. As with the commander_id, it can be
        a string or an integer
    consumer_id
        The actor that is consuming this command. Normally this is our own
        actor but if we are commanding another actor ``consumer_id`` will
        be the destination actor.
    actor
        The actor instance associated to this command.
    parent
        Another `.BaseCommand` object that is issuing this subcommand.
        Messages emitted by the command will use the parent ``command_id``.
    reply_callback
        A callback that gets called when a client command receives a reply from the
        actor.
    status_callback
        A function to call when the status changes.
    call_now
        Whether to call ``status_callback`` when initialising the command.
    default_keyword
        The keyword to use when writing a message that does not specify a
        keyword.
    silent
        A silent command will call the actor ``write`` method with ``silent=True``,
        which will update the internal model and record all the output replies but
        will not write them to the users.
    internal
        A silent command will call the actor ``write`` method with ``internal=True``,
        which will instruct the actor to add the internal flag to the header of the
        reply.
    write_to_log
        Whether to write replies to the log. Defaults to yes but it may be useful
        to prevent large repetitive replies cluttering the log.
    time_limit
        Time out the command if it has been running for this long.
    loop
        The event loop.

    """

    def __init__(
        self,
        commander_id: Union[int, str, None] = None,
        command_id: Union[int, str] = 0,
        consumer_id: Union[int, str] = 0,
        actor: Optional[Actor_co] = None,
        parent: Optional[BaseCommand[Actor_co, Future_co]] = None,
        reply_callback: Optional[Callable[[Any], None]] = None,
        status_callback: Optional[Callable[[CommandStatus], Any]] = None,
        call_now: bool = False,
        default_keyword: str = "text",
        silent: bool = False,
        internal: bool = False,
        write_to_log: bool = True,
        time_limit: float | None = None,
        loop: Optional[asyncio.AbstractEventLoop] = None,
    ):
        self.commander_id = commander_id
        self.consumer_id = consumer_id
        self.command_id = command_id

        # Casting here so that we can type Command[SomeActor] and not have to
        # assert command.actor every time.
        self.actor = cast(Actor_co, actor)
        self.parent = parent

        self.silent = silent
        self.internal = internal
        self.write_to_log = write_to_log

        self._reply_callback = reply_callback

        self.default_keyword = default_keyword
        self.loop = loop or asyncio.get_event_loop()

        #: The click context, if the click parser is used.
        self.context: click.Context | None = None

        #: A list of replies this command has received. The type of
        #: reply object depends on the actor or client issuing the command.
        self.replies = ReplyList([])

        asyncio.Future.__init__(self)

        self._status: CommandStatus
        StatusMixIn.__init__(
            self,
            CommandStatus,
            initial_status=CommandStatus.READY,
            callback_func=status_callback,
            call_now=call_now,
        )

        self._timer_handler = None
        if time_limit:
            self._timer_handler = asyncio.get_running_loop().call_later(
                time_limit,
                self.set_status,
                CommandStatus.TIMEDOUT,
            )

    @property
    def status(self) -> CommandStatus:
        """Returns the status."""

        return self._status

    @status.setter
    def status(self, status: CommandStatus):
        """Sets the status. A message is output to the users.

        This setter calls `.set_status` with an empty message to the users.

        Parameters
        ----------
        status
            The status to set, either as a `CommandStatus` value or the
            integer associated with the maskbit. If ``value`` is a string,
            loops over the bits in `CommandStatus` and assigns the one whose
            name matches.
        """

        self.set_status(status)

    def set_status(
        self,
        status: Union[CommandStatus, str],
        message: Dict[str, Any] | str | None = None,
        silent: bool = False,
        **kwargs,
    ) -> BaseCommand:
        """Same as `.status` but allows to specify a message to the users."""

        assert self.status

        # Don't do anything if the command is done. This means the command
        # finished before the timeout happened.
        if status == CommandStatus.TIMEDOUT and self.done():
            return self

        if self.status.is_done:
            raw_command_string = getattr(self, "raw_command_string", "NA")
            warnings.warn(
                f"{raw_command_string}: cannot modify a "
                f"done command with status {status!r}.",
                CluWarning,
            )
            return self

        if isinstance(status, str):
            for bit in self.flags:
                assert bit.name
                if status.lower() == bit.name.lower():
                    status = bit
                    break

        status = self.flags(status)

        try:
            is_flag = status in self.flags
            if not is_flag:
                raise TypeError()
        except TypeError:
            raise TypeError(f"Status {status!r} is not a valid command status.")

        if status != self._status:
            status_code = status.code
            if status_code is None:
                raise ValueError(f"Invalid status code {status_code!r}.")

            if isinstance(message, str) and status_code in ["f", "e"]:
                message = {"error": message}

            if self.actor and not silent:
                self.write(status_code, message, **kwargs)

            self._status = status

            self.do_callbacks()

            # If the command is done, set the result of the future.
            if self._status.is_done and not self.done():
                self.set_result(self)  # type: ignore
                if self._timer_handler:
                    self._timer_handler.cancel()

            # Set the status watcher
            if self.watcher is not None:
                self.watcher.set()

        return self

    def finish(self, *args, **kwargs):
        """Convenience method to mark a command `~.CommandStatus.DONE`."""

        self.set_status(CommandStatus.DONE, *args, **kwargs)
        return self

    def fail(self, *args, **kwargs):
        """Convenience method to mark a command `~.CommandStatus.FAILED`."""

        self.set_status(CommandStatus.FAILED, *args, **kwargs)
        return self

    def debug(self, *args, **kwargs):
        """Writes a debug-level message."""

        self.write("d", *args, **kwargs)

    def info(self, *args, **kwargs):
        """Writes an info-level message."""

        self.write("i", *args, **kwargs)

    def warning(self, *args, **kwargs):
        """Writes a warning-level message."""

        self.write("w", *args, **kwargs)

    def error(self, *args, **kwargs):
        """Writes an error-level message (does not fail the command)."""

        self.write("e", *args, **kwargs)

    def write(
        self,
        message_code: clu.base.MessageCode | str | int = clu.base.MessageCode.INFO,
        message: Optional[Union[Dict[str, Any], str]] = None,
        broadcast: bool = False,
        **kwargs,
    ):
        """Writes to the user(s).

        Parameters
        ----------
        message_code
            The message code (e.g., ``'i'`` or ``':'``).
        message
            The text to be output. If `None`, only the code will be written.
        """

        if message is None:
            message = {}
        elif isinstance(message, dict):
            pass
        elif isinstance(message, str):
            keyword = "error" if message_code in ["f", "e"] else self.default_keyword
            message = {keyword: message}
        elif isinstance(message, Exception):
            message = {"error": message}
        else:
            raise ValueError(f"invalid message {message!r}")

        if not self.actor:
            raise clu.CommandError(
                "An actor has not been defined for "
                "this command. Cannot write to users."
            )

        command = self if not self.parent else self.parent

        # If the message code is an integer, interpret that as if it's a logging
        # level and translate it to SDSS codes.
        if isinstance(message_code, int):
            if message_code == logging.DEBUG:
                message_code = "d"
            elif message_code == logging.INFO:
                message_code = "i"
            elif message_code == logging.WARNING:
                message_code = "w"
            elif message_code == logging.ERROR:
                message_code = "e"
            else:
                raise ValueError(f"Invalid message code {message_code}.")

        message_code = clu.base.MessageCode(message_code)

        # If the parent has a command, do not output : or f since it would
        # confuse the stream and potentially Tron.
        if self.parent:
            if message_code == clu.base.MessageCode.STARTED:
                # The parent is already running and > never includes a message.
                return
            if message_code == clu.base.MessageCode.DONE:
                message_code = clu.base.MessageCode.INFO
                if kwargs == {} and (message == {} or not message):
                    return
            elif message_code == clu.base.MessageCode.FAILED:
                message_code = clu.base.MessageCode.ERROR

        internal = kwargs.pop("internal", self.internal)
        write_to_log = kwargs.pop("write_to_log", self.write_to_log)

        self.actor.write(
            message_code,
            message=message,
            command=command,
            broadcast=broadcast,
            silent=self.silent,
            internal=internal,
            write_to_log=write_to_log,
            **kwargs,
        )

    def send_command(
        self,
        target: str,
        command_string: str,
        new_command: bool = True,
        *args,
        **kwargs,
    ) -> BaseCommand[Actor_co, BaseCommand] | Awaitable[BaseCommand]:
        """Sends a command to an actor using the commander ID of this command."""

        if self.actor is None:
            raise CommandError("The actor need to be defined to send commands.")

        return self.actor.send_command(
            target,
            command_string,
            *args,
            command=self if new_command is False else None,
            **kwargs,
        )


class Command(BaseCommand[Actor_co, "Command"]):
    """A command from a user.

    Parameters
    ----------
    command_string
        The string that defines the body of the command.
    transport
        The TCP transport associated with this command (only relevant
        for `.LegacyActor` commands).
    """

    def __init__(
        self,
        command_string: str = "",
        transport: Optional[Any] = None,
        **kwargs,
    ):
        BaseCommand.__init__(self, **kwargs)

        #: The raw command string.
        self.raw_command_string = command_string

        #: The body of the command.
        self.body = command_string

        if not self.actor and self.parent:
            self.actor = self.parent.actor

        #: The `~click.Context` running this command. Only relevant if
        #: using the built-in click-based parser.
        self.ctx = None

        self.transport = transport

    def child_command(self, command_string):
        """Starts a sub-command on the actor currently running this command.

        The practical effect is to run another of the same actor's commands as
        if it were part of the current `.Command`.

        """

        return Command(
            command_string,
            actor=self.actor,
            commander_id=self.actor.name,
            parent=self,
        ).parse()

    def parse(self) -> Command:
        """Parses the command."""

        if not isinstance(self.actor, clu.base.BaseActor):
            raise clu.CluError("the actor is not defined. Cannot parse command.")

        self.actor.parse_command(self)

        return self

    def __str__(self):
        return (
            f"<Command (commander_id={self.commander_id!r}, "
            f"command_id={self.command_id!r}, body={self.body!r})>"
        )


def parse_legacy_command(command_string: str) -> Tuple[Union[str, None], int, str]:
    """Parses a command received by a legacy actor.

    Parameters
    ----------
    command_string
        The command to parse, including an optional header.

    Returns
    -------
    command_id, command_body
        The command ID, and the command body parsed from the command
        string.

    """

    _HEADER_BODY_RE = re.compile(
        r"(?:([a-z0-9]*\.[a-z0-9_\.]+)\s+)?"
        r"(?:(\d+)(?:\s+\d+)?\s+)?"
        r"(?:([a-z_].*|--help))?$",
        re.IGNORECASE,
    )

    command_match = _HEADER_BODY_RE.match(command_string)
    if not command_match:
        raise clu.CommandError(f"Could not parse command {command_string!r}")

    commander, command_id_str, command_body = command_match.groups()

    if command_id_str:
        command_id = int(command_id_str)
    else:
        command_id = 0

    command_body = command_body.strip()

    return commander, command_id, command_body


class TimedCommandList(list):
    """A list of `.TimedCommand` objects that will be executed on a loop.

    Parameters
    ----------
    actor
        The actor in which the commands are to be run.
    resolution
        In seconds, how frequently to check if any of the `.TimedCommand` must
        be executed.

    """

    def __init__(
        self,
        actor: clu.base.BaseActor,
        resolution=0.5,
        loop: Optional[asyncio.AbstractEventLoop] = None,
    ):
        self.resolution = resolution
        self.actor = actor
        self.loop = loop or asyncio.get_event_loop()

        self._task: Optional[asyncio.Task] = None

        list.__init__(self, [])

    def add_command(self, command_string: str, **kwargs):
        """Adds a new `.TimedCommand`."""

        self.append(TimedCommand(command_string, **kwargs))

    async def poller(self):
        """The polling loop."""

        current_time = time.time()
        first_time = True

        while True:
            for timed_command in self:
                elapsed = current_time - timed_command.last_run
                if first_time or elapsed > timed_command.delay:
                    timed_command_task = self.loop.create_task(
                        timed_command.run(self.actor)
                    )
                    timed_command_task.add_done_callback(timed_command.done)

            first_time = False

            self._sleep_task = self.loop.create_task(asyncio.sleep(self.resolution))

            await self._sleep_task
            current_time += self.resolution

    def start(self) -> TimedCommandList:
        """Starts the loop."""

        if self.running:
            raise RuntimeError("poller is already running.")

        self._task = self.loop.create_task(self.poller())

        return self

    async def stop(self):
        """Cancel the poller."""

        if not self.running:
            return

        if self._task:
            self._task.cancel()
            with suppress(asyncio.CancelledError):
                await self._task

    @property
    def running(self) -> bool:
        """Returns `True` if the poller is running."""

        if self._task and not self._task.cancelled():
            return True

        return False


class TimedCommand(object):
    """A command to be executed on a loop.

    Parameters
    ----------
    command_string
        The command string to run.
    delay
        How many seconds to wait between repeated calls.
    first_silent
        Runs the command in silent mode the first one (useful to internally update the
        model).

    """

    def __init__(self, command_string: str, delay: float = 1, first_silent=False):
        self.command_string = command_string
        self.delay = delay

        self.last_run = 0.0
        self.is_running = False

        self.first_silent = first_silent

    async def run(self, actor: clu.base.BaseActor):
        """Run the command."""

        if self.is_running:
            return

        silent = True if self.first_silent and self.last_run == 0.0 else False

        self.is_running = True

        await Command(
            self.command_string,
            actor=actor,
            commander_id=f".{actor.name}",
            silent=silent,
        ).parse()

    def done(self, task):
        """Marks the execution of a command."""

        self.last_run = time.time()
        self.is_running = False


class FakeCommand(BaseCommand):
    """A fake command that output to a logger."""

    def __init__(self, log: logging.Logger, actor=None):
        self.log = log

        super().__init__(actor=actor)

    def write(
        self,
        message_code: str = "i",
        message: Optional[Union[Dict[str, Any], str]] = None,
        **kwargs,
    ):
        if message_code == "d":
            level = logging.DEBUG
        elif message_code == "i":
            level = logging.INFO
        elif message_code == "w":
            level = logging.WARNING
        elif message_code in ["f", "e"]:
            level = logging.ERROR
        else:
            return

        if message is None:
            message = {}
        elif isinstance(message, dict):
            pass
        elif isinstance(message, str):
            keyword = "error" if message_code in ["f", "e"] else self.default_keyword
            message = {keyword: message}
        elif isinstance(message, Exception):
            message = {"error": message}
        else:
            raise ValueError(f"invalid message {message!r}")

        message.update(kwargs)

        self.log.log(level, str(message))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2019-05-10
# @Filename: tron.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import logging
import re
import time
import warnings
from threading import Lock

from typing import Any, Callable, List, Optional

import clu.base
from clu.command import Command, CommandStatus
from clu.exceptions import CluWarning
from clu.model import CluModel, Property
from clu.protocol import ReconnectingTCPClientProtocol

from .types.keys import Key, KeysDictionary
from .types.messages import Keyword, Reply
from .types.parser import ParseError, ReplyParser


__all__ = ["TronConnection", "TronModel", "TronKey"]


class TronKey(Property):
    """A Tron model key with callbacks.

    Similar to `.Property` but stores the original key with the keyword
    datamodel.
    """

    def __init__(
        self,
        name: str,
        key: Key,
        keyword: Optional[Keyword] = None,
        model: Optional[TronModel] = None,
        callback: Optional[Callable[[TronKey], Any]] = None,
    ):
        initial_value = [None] * len(key.typedValues.vtypes)
        super().__init__(name, value=initial_value, model=model, callback=callback)

        self.key = key
        self.keyword = None

        self.update_keyword(keyword)

    def update_keyword(self, keyword: Optional[Keyword]):
        """Updates the keyword and value."""

        if keyword is None:
            return

        self.keyword = keyword
        self.value = [value.native for value in keyword.values]

    def __getitem__(self, sl):
        return self.value.__getitem__(sl)

    def __len__(self):
        return len(self.value)


class TronModel(CluModel[TronKey]):
    """A JSON-compliant model for actor keywords.

    Parameters
    ----------
    keydict
        A dictionary of keys that define the datamodel.
    callback
        A function or coroutine to call when the datamodel changes. The
        function is called with the instance of `.TronModel` and the modified keyword.
        If the callback is a coroutine, it is scheduled as a task.

    """

    def __init__(
        self,
        keydict: KeysDictionary,
        callback: Optional[Callable[[TronModel], Any]] = None,
    ):
        super().__init__(keydict.name, callback=callback)

        self.keydict = keydict

        for key in self.keydict.keys:
            key = self.keydict.keys[key]
            self[key.name] = TronKey(key.name, key, model=self)

        self._lock = Lock()

    def reload(self):
        """Reloads the model. Clears callbacks."""

        model = self.keydict.name
        keydict = KeysDictionary.load(model)

        self.__init__(keydict)

    def parse_reply(self, reply):
        """Parses a reply and updates the datamodel."""

        parsed: dict[str, Any] = {}

        with self._lock:
            for reply_key in reply.keywords:
                self.last_seen = time.time()

                key_name = reply_key.name.lower()
                if key_name not in self.keydict:
                    warnings.warn(
                        f"Cannot parse unknown keyword {self.name}.{reply_key.name}.",
                        CluWarning,
                    )
                    continue

                # When parsed the values in reply_key are string. After consuming
                # it with the Key, the values become typed values.
                result = self.keydict.keys[key_name].consume(reply_key)

                if not result:
                    warnings.warn(
                        f"Failed parsing keyword {self.name}.{reply_key.name}.",
                        CluWarning,
                    )
                    continue

                self[key_name].update_keyword(reply_key)
                parsed[key_name] = self[key_name].value.copy()

                self.notify(self.flatten(), self[key_name].copy())

        return parsed


class TronLoggingFilter(logging.Filter):
    """Logs issues with the Tron parser only to the file logger."""

    def filter(self, record):
        return not record.getMessage().startswith("Failed parsing reply")


class TronClientProtocol(ReconnectingTCPClientProtocol):
    """A reconnecting protocol for the Tron connection."""

    def __init__(self, on_received, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._on_received = on_received
        self.transport: asyncio.Transport | None = None

    def data_received(self, data):
        asyncio.get_event_loop().call_soon(self._on_received, data)

    def connection_made(self, transport: asyncio.Transport):
        self.transport = transport


class TronConnection(clu.base.BaseClient):
    """Allows to send commands to Tron and manages the feed of replies.

    Parameters
    ----------
    commander
        The name of the commander that will send commands to Tron. Must be
        in the form ``program.user``, for example ``foo.bar``.
    host
        The host on which Tron is running.
    port
        The port on which Tron is running.
    models
        A list of strings with the actors whose models will be tracked.
    kwargs
        Arguments to be passed to `.BaseClient`.
    """

    def __init__(
        self,
        commander: str,
        host: str,
        port: int = 6093,
        name: str = "tron",
        models: List[str] = [],
        **kwargs,
    ):
        super().__init__(name, **kwargs)

        self.commander = commander
        if not re.match(r"[A-Za-z_]+\.[A-Za-z_]+", self.commander):
            raise ValueError("Invalid commander format.")

        self.host = host
        self.port = port

        self._mid = 1

        models = models or []

        #: dict: The `KeysDictionary` associated with each actor to track.
        self.keyword_dicts = {model: KeysDictionary.load(model) for model in models}

        #: dict: The model and values of each actor being tracked.
        self.models = {model: TronModel(self.keyword_dicts[model]) for model in models}

        self.rparser: Any = ReplyParser()

        self.protocol: TronClientProtocol | None = None

        self.running_commands = {}

        self.buffer = b""

        # We want to log problems with the Tron parser, but not to the console.
        if self.log.sh:
            self.log.sh.addFilter(TronLoggingFilter())

    async def start(self, get_keys=True):
        """Starts the connection to Tron.

        Parameters
        ----------
        get_keys : bool
            If `True`, gets all the keys in the models.
        """

        self.set_loop_exception_handler()

        self.protocol = TronClientProtocol(
            self._handle_reply,
            self.host,
            self.port,
        )
        await self.protocol._connect()

        if not self.protocol.connected:
            self.log.error(f"Failed connecting to Tron on ({self.host}, {self.port})")
            return

        if get_keys:
            asyncio.create_task(self.get_keys())

        return self

    def stop(self):
        """Closes the connection."""

        if self.protocol:
            self.protocol.stop_trying()

        if self.protocol and self.protocol.transport:
            self.protocol.transport.close()

        self.protocol = None

    def connected(self):
        """Checks whether the client is connected."""

        if self.protocol and self.protocol.connected:
            return True

        return False

    async def run_forever(self):  # pragma: no cover
        assert self.protocol and self.connected()

        # Keep alive until the connection is closed.
        while True:
            await asyncio.sleep(1)
            if self.protocol is None:
                return

    def send_command(
        self,
        target,
        command_string,
        *args,
        commander=None,
        mid=None,
        callback: Optional[Callable[[Reply], None]] = None,
        time_limit: Optional[float] = None,
    ):
        """Sends a command through the hub.

        Parameters
        ----------
        target
            The actor to command.
        command_string
            The command to send.
        args
            Arguments to concatenate to the command string.
        commander
            The actor or client sending the command. The format for Tron is
            "commander message_id target command" where commander needs to
            start with a letter and have a program and a user joined by a dot.
            Otherwise the command will be accepted but the reply will fail
            to parse. If ``commander=None``, the instance ``commander`` value
            will be used.
        mid
            The message id. If `None`, a sequentially increasing value will
            be used. You should not specify a ``mid`` unless you really know
            what you're doing.
        callback
            A callback to invoke with each reply received from the actor.
        time_limit
            A delay after which the command is marked as timed out and done.

        Examples
        --------
        These two are equivalent ::

            >>> tron.send_command('my_actor', 'do_something --now')
            >>> tron.send_command('my_actor', 'do_something', '--now')

        """

        assert self.protocol and self.protocol.transport and self.connected()

        mid = mid or self._mid

        # The mid must be a 32-bit unsigned number.
        if mid >= 2**32:
            self._mid = mid = mid % 2**32

        if len(args) > 0:
            command_string += " " + " ".join(map(str, args))

        commander = commander or self.commander

        command_string = f"{commander} {mid} {target} {command_string}\n"

        command = Command(
            command_string=command_string,
            reply_callback=callback,
            time_limit=time_limit,
            commander_id=commander,
        )
        self.running_commands[mid] = command

        self.protocol.transport.write(command_string.encode())

        self._mid += 1

        return command

    async def get_keys(self):
        """Gets all the keys for the models being tracked."""

        # Number of keys to be requested at once
        n_keys = 10

        for model in self.models.values():
            actor = model.name
            keys = [key.lower() for key in model]

            for ii in range(0, len(keys), n_keys):
                keys_to_request = keys[ii : ii + n_keys]

                if len(keys_to_request) == 0:
                    break

                keys_joined = " ".join(keys_to_request)

                command_string = f"getFor={actor} {keys_joined}"

                self.send_command("keys", command_string)

    def _handle_reply(self, data: bytes):
        """Tracks new replies from Tron and updates the model."""

        self.buffer += data

        lines = self.buffer.splitlines()
        if not self.buffer.endswith(b"\n"):
            self.buffer = lines.pop()
        else:
            self.buffer = b""

        for line in lines:
            try:
                # Do not strip here or that will cause parsing problems.
                line = line.decode()
                reply: Reply = self.rparser.parse(line)
            except ParseError:
                self.log.warning(f"Failed parsing reply '{line.strip()}'.")
                continue

            reply_actor = reply.header.actor
            reply_commander_id = reply.header.cmdrName

            # The keys command returns keywords as if from the actor
            # keys_<actor> (e.g. keys_tcc).
            if reply_actor.startswith("keys_"):
                reply_actor = reply_actor.split("_")[1]

            parsed_data = {}
            if reply_actor in self.models:
                try:
                    parsed_data = self.models[reply_actor].parse_reply(reply)
                except ParseError as ee:
                    self.log.warning(
                        f"Failed parsing reply {reply!r} with error: {ee!s}"
                    )
            else:
                # Fallback in case the actor of the reply is not in the models.
                # In this case the values will be strings.
                parsed_data = {kw.name: kw.values for kw in reply.keywords}

            mid = reply.header.commandId
            status = CommandStatus.code_to_status(reply.header.code.lower())

            if mid in self.running_commands:
                # We may be receiving messages from a command with the same MID
                # that's not managed by this instance of the tron client, so we
                # also check the commander.
                if self.running_commands[mid].commander_id == reply_commander_id:
                    self.running_commands[mid].replies.append(
                        clu.base.Reply(
                            message={k: v for k, v in parsed_data.items()},
                            message_code=reply.header.code.lower(),
                            command=self.running_commands[mid],
                            validated=True,
                            keywords=reply.keywords,
                        )
                    )
                    self.running_commands[mid].set_status(status)
                    if self.running_commands[mid]._reply_callback is not None:
                        self.running_commands[mid]._reply_callback(reply)
                    if status.is_done:
                        self.running_commands.pop(mid)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2019-10-05
# @Filename: testing.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import json
import re
import types
import unittest.mock

from typing import TYPE_CHECKING, Any, Dict, List, Optional, TypeVar

import aio_pika
from aiormq.abc import DeliveredMessage
from pamqp.commands import Basic
from pamqp.header import ContentHeader

import clu
from clu.actor import AMQPBaseActor, TCPBaseActor
from clu.command import Command
from clu.legacy.actor import LegacyActor


if TYPE_CHECKING:
    from pamqp.common import FieldTable


__all__ = ["MockReply", "MockReplyList", "setup_test_actor"]


class MockedActor(TCPBaseActor, LegacyActor, AMQPBaseActor):
    invoke_mock_command: Any
    mock_replies: List[MockReply]


T = TypeVar("T", bound=MockedActor)


class MockReply(dict):
    """Stores a reply written to a transport.

    The data of the message is stored as part of the dictionary.

    Parameters
    ----------
    user_id
        The user ID of the client to which the reply was sent.
    command_id
        The command ID of the command that produced this reply.
    flag
        The message type flag.
    data
        The payload of the message.
    """

    def __init__(
        self,
        user_id: int | str | None,
        command_id: int | str | None,
        flag: str,
        data: Dict[str, Any] = {},
    ):
        self.command_id = command_id
        self.user_id = user_id
        self.flag = flag

        dict.__init__(self, data)

    def __repr__(self):
        return (
            f"<MockReply ({self.user_id} {self.command_id} "
            f"{self.flag} {super().__repr__()})>"
        )


class MockReplyList(list):
    """Stores replies as `.MockReply` objects."""

    LEGACY_REPLY_PATTERN = re.compile(
        r"([0-9]+)\s+([0-9]+)\s+((?:[a-z]|\:|\>|\!))\s+(.*)"
    )

    def __init__(self, actor):
        self.actor = actor

        list.__init__(self)

    def parse_reply(
        self,
        reply: bytes | str | aio_pika.Message,
        routing_key: Optional[str] = None,
    ):
        """Parses a reply and construct a `.MockReply`, which is appended."""

        if isinstance(reply, bytes):
            reply = reply.decode()

        if issubclass(self.actor.__class__, clu.LegacyActor):
            assert isinstance(reply, str)
            match = self.LEGACY_REPLY_PATTERN.match(reply)
            if not match:
                return

            user_id, command_id, flag, keywords_raw = match.groups()

            user_id = int(user_id)
            command_id = int(command_id)

            data = {}
            for keyword_raw in keywords_raw.split(";"):
                if keyword_raw.strip() == "":
                    continue
                if "=" in keyword_raw:
                    name, value = keyword_raw.split("=", maxsplit=1)
                else:
                    name = keyword_raw
                    value = ""

                data[name.strip()] = value.strip()

        elif issubclass(self.actor.__class__, clu.JSONActor):
            assert isinstance(reply, str)
            reply_dict: Dict[str, Any] = json.loads(reply)

            header = reply_dict["header"]
            user_id = header.pop("commander_id", None)
            command_id = header.pop("command_id", None)
            flag = header.pop("message_code", "d")

            data = reply_dict["data"]

        elif issubclass(self.actor.__class__, clu.AMQPActor):
            assert isinstance(reply, aio_pika.Message)

            header = reply.headers

            user_id = header.get("commander_id", None)
            command_id = header.get("command_id", None)
            flag = header.get("message_code", "d")

            data = json.loads(reply.body.decode())

        else:
            raise RuntimeError("This type of actor is not supported")

        assert isinstance(user_id, (int, str)) or user_id is None
        assert isinstance(command_id, (int, str)) or command_id is None
        assert isinstance(flag, str)

        list.append(self, MockReply(user_id, command_id, flag, data))

    def clear(self):
        list.__init__(self)

    def __contains__(self, m):
        return any([m in reply[kw] for reply in self for kw in reply.keys()])


async def setup_test_actor(actor: T, user_id: int = 666) -> T:
    """Setups an actor for testing, mocking the client transport.

    Takes an ``actor`` and modifies it in two ways:

    - Adds a ``invoke_mock_command`` method to it that allows to submit a
      command string as if it had been received from a transport.

    - Mocks a client transport with ``user_id`` that is connected to the actor.
      Messages written to the transport are stored as `.MockReply` in a
      `.MockReplyList` that is accessible via a new ``actor.mock_replies``
      attribute.

    The actor is modified in place and returned.
    """

    if not issubclass(actor.__class__, (clu.LegacyActor, clu.JSONActor, clu.AMQPActor)):
        raise RuntimeError("setup_test_actor is not implemented for this type of actor")

    def invoke_mock_command(self, command_str, command_id=0):
        if issubclass(actor.__class__, (clu.LegacyActor, clu.JSONActor)):
            if isinstance(command_str, str):
                command_str = command_str.encode("utf-8")
            full_command = f" {command_id} ".encode("utf-8") + command_str
            return self.new_command(actor.transports[user_id], full_command)
        elif issubclass(actor.__class__, clu.AMQPActor):
            command_id = str(command_id)
            headers: FieldTable = {
                "command_id": command_id,
                "commander_id": "mock_test_client",
            }
            header = ContentHeader(
                properties=Basic.Properties(
                    content_type="text/json",
                    headers=headers,
                )
            )
            message_body = {"command_string": command_str}
            message = aio_pika.IncomingMessage(
                DeliveredMessage(
                    Basic.Deliver(),
                    header,
                    json.dumps(message_body).encode(),
                    None,  # type: ignore
                ),
            )
            return self.new_command(message, ack=False)

    actor.start = unittest.mock.AsyncMock(return_value=actor)

    # Adds an invoke_mock_command method.
    # We use types.MethodType to bind a method to an existing instance
    # (see http://bit.ly/35buk1m)
    actor.invoke_mock_command = types.MethodType(invoke_mock_command, actor)

    # Mocks a user transport and stores the replies in a MockReplyListobject
    actor.mock_replies = MockReplyList(actor)

    if issubclass(actor.__class__, (clu.LegacyActor, clu.JSONActor)):
        mock_transport = unittest.mock.MagicMock(spec=asyncio.Transport)
        mock_transport.user_id = user_id
        mock_transport.write.side_effect = actor.mock_replies.parse_reply
        actor.transports[user_id] = mock_transport
    elif issubclass(actor.__class__, clu.AMQPActor):
        assert actor.connection
        actor.connection.exchange = unittest.mock.MagicMock()
        actor.connection.exchange.publish = unittest.mock.AsyncMock(
            side_effect=actor.mock_replies.parse_reply
        )

    await actor.start()

    return actor


class TestCommand(Command):  # pragma: no cover
    """A `.Command` that can be reset."""

    def reset(self):
        """Resets the command."""

        self._status = self.flags.READY

        if self.watcher:
            self.watcher.clear()
        self.watcher = None

        asyncio.Future.__init__(self)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2019-05-20
# @Filename: __main__.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import asyncio

import aio_pika
import click
from click_default_group import DefaultGroup

import clu


color_codes = {
    ">": "lightblue",
    "i": "lightblue",
    "d": "gray",
    "w": "yellow",
    "e": "LightCoral",
    "f": "red",
    ":": "green",
}


class ShellClient(clu.AMQPClient):
    """A shell client."""

    actors: tuple[str, ...] = ()
    indent = False
    show_time = True
    ignore_broadcasts = False
    internal = False
    all_ = False

    async def handle_reply(self, message: aio_pika.IncomingMessage):
        """Prints the formatted reply."""

        import datetime
        import json
        from datetime import timezone

        import prompt_toolkit
        import pygments.lexers
        from prompt_toolkit import print_formatted_text as print
        from prompt_toolkit.formatted_text import PygmentsTokens
        from prompt_toolkit.styles import style_from_pygments_cls
        from pygments.styles import STYLE_MAP, get_style_by_name

        reply = await super().handle_reply(message)

        if reply is None or (self.internal is False and reply.internal is True):
            return

        headers = reply.info.get("headers", {})
        commander_id = headers.get("commander_id", None)
        if commander_id:
            commander_id = str(commander_id).split(".")[0]

        routing_key = message.routing_key
        is_broadcast = routing_key == "reply.broadcast" or reply.command_id is None

        if (
            commander_id
            and commander_id != self.name
            and not is_broadcast
            and not self.all_
        ):
            return

        if self.ignore_broadcasts and is_broadcast:
            return

        message_code = reply.message_code
        sender = headers.get("sender", "")

        if is_broadcast and (len(self.actors) > 0 and sender not in self.actors):
            return

        message_code_esc = message_code if message_code != ">" else "&gt;"

        message_code_formatted = prompt_toolkit.formatted_text.HTML(
            f'<style font-weight="bold" '
            f'fg="{color_codes[message_code]}">{message_code_esc}</style>'
        )

        body = reply.body

        indent = 4 if self.indent is True else None
        body_tokens = list(
            pygments.lex(
                json.dumps(body, indent=indent),
                lexer=pygments.lexers.JsonLexer(),  # type: ignore
            )
        )

        print_chunks = []

        if self.show_time:
            time = datetime.datetime.now(timezone.utc).isoformat().split("T")[1]
            time = time[0:12]  # Milliseconds
            print_chunks.append(
                prompt_toolkit.formatted_text.HTML(f'<style fg="Gray">{time}</style>'),
            )

        if sender:
            print_chunks.append(
                prompt_toolkit.formatted_text.HTML(
                    f'<style fg="magenta">{sender}</style>'
                ),
            )

        if message_code:
            print_chunks.append(message_code_formatted)

        style_name = "solarized-dark" if "solarized-dark" in STYLE_MAP else "default"
        style = style_from_pygments_cls(get_style_by_name(style_name))  # type: ignore

        if body:
            print_chunks.append(PygmentsTokens(body_tokens))
            print(*print_chunks, style=style, end="")
        else:
            print(*print_chunks, flush=True)  # Newline


async def shell_client_prompt(
    actors: tuple[str, ...] = (),
    url: str | None = None,
    user: str = "guest",
    password: str = "guest",
    host: str = "127.0.0.1",
    port: int = 5672,
    all_: bool = False,
    indent=True,
    show_time=True,
    ignore_broadcasts=False,
    internal=False,
):
    import os
    import uuid

    import prompt_toolkit
    from prompt_toolkit.history import FileHistory
    from prompt_toolkit.patch_stdout import patch_stdout

    # Give each client a unique name to ensure the queues are unique.
    uid = str(uuid.uuid4()).split("-")[0]

    client = await ShellClient(
        f"shell_client_{uid}",
        url=url,
        user=user,
        password=password,
        host=host,
        port=port,
        log=None,
    ).start()

    client.actors = actors
    client.indent = indent
    client.show_time = show_time
    client.ignore_broadcasts = ignore_broadcasts
    client.internal = internal
    client.all_ = all_

    history = FileHistory(os.path.expanduser("~/.clu_history"))

    session = prompt_toolkit.PromptSession("", history=history)

    while True:
        with patch_stdout():
            try:
                text = await session.prompt_async("")
            except KeyboardInterrupt:
                break
            except EOFError:
                break
            else:
                text = text.strip()
                if text.startswith("quit"):
                    break
                elif text == "":
                    continue
                else:
                    chunks = text.split()
                    if len(chunks) < 2:
                        print(f"Invalid command {text!r}")
                        continue

                    actor = chunks[0]
                    command_string = " ".join(chunks[1:])

                    await client.send_command(
                        actor,
                        command_string,
                        await_command=False,
                    )


@click.group(cls=DefaultGroup, default="cli", default_if_no_args=True, name="clu")
def clu_cli():
    """CLU command line interface."""

    pass


@clu_cli.command(name="cli")
@click.argument(
    "ACTORS",
    type=str,
    required=False,
    nargs=-1,
)
@click.option(
    "--url",
    type=str,
    help="AMQP RFC3986 formatted broker address.",
)
@click.option(
    "--user",
    "-U",
    type=str,
    show_default=True,
    default="guest",
    help="The AMQP username.",
)
@click.option(
    "--password",
    "-P",
    type=str,
    show_default=True,
    default="guest",
    help="The AMQP password.",
)
@click.option(
    "--host",
    "-H",
    type=str,
    show_default=True,
    default="localhost",
    help="The host running the AMQP server.",
)
@click.option(
    "--port",
    "-p",
    type=int,
    show_default=True,
    default=5672,
    help="The port on which the server is running",
)
@click.option(
    "--all",
    "-a",
    "all_",
    is_flag=True,
    help="Print all messages, not only those replying to commands sent from this CLI.",
)
@click.option(
    "--no-indent",
    "-n",
    is_flag=True,
    default=False,
    help="Do not indent the output JSONs.",
)
@click.option(
    "--ignore-broadcasts",
    "-b",
    is_flag=True,
    default=False,
    help="Only show replies to client commands.",
)
@click.option(
    "--no-time",
    "-t",
    is_flag=True,
    default=False,
    help="Do not show the message time.",
)
@click.option(
    "--internal",
    is_flag=True,
    help="Show internal messages.",
)
def cli(
    actors: tuple[str, ...] = (),
    url: str | None = None,
    user: str = "guest",
    password: str = "guest",
    host: str = "localhost",
    port: int = 5672,
    all_: bool = False,
    no_indent=False,
    no_time=False,
    ignore_broadcasts=False,
    internal=False,
):
    """Runs the AMQP command line interpreter."""

    shell_task = asyncio.get_event_loop().create_task(
        shell_client_prompt(
            actors=actors,
            url=url,
            user=user,
            password=password,
            host=host,
            port=port,
            all_=all_,
            indent=not no_indent,
            show_time=not no_time,
            ignore_broadcasts=ignore_broadcasts,
            internal=internal,
        )
    )
    asyncio.get_event_loop().run_until_complete(shell_task)


@clu_cli.command()
@click.option(
    "--port",
    "-p",
    default=9876,
    show_default=True,
    type=int,
    help="The port on which to serve the websocket",
)
@click.option(
    "--rabbitmq-host",
    "-H",
    type=str,
    show_default=True,
    default="localhost",
    help="The host running the AMQP server.",
)
@click.option(
    "--rabbitmq-port",
    "-P",
    type=int,
    show_default=True,
    default=5672,
    help="The port on which the AMQP server is running.",
)
def websocket(port=9876, rabbitmq_host="localhost", rabbitmq_port=5672):
    """Launches a websocket server connected to a RabbitMQ exchange."""

    try:
        from clu.websocket import WebsocketServer
    except ImportError:
        click.echo("websockets package is needed. Try pip install sdss-clu[websocket].")
        return

    async def _async_helper():
        ws = WebsocketServer(wport=port, host=rabbitmq_host, port=rabbitmq_port)
        await ws.start()

        # Run forever.
        await asyncio.Future()

    asyncio.run(_async_helper())


def main():
    clu_cli(obj={})


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2019-05-14
# @Filename: device.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import contextlib

from typing import Any, Callable, Optional, TypeVar

from .protocol import open_connection
from .tools import CallbackMixIn


__all__ = ["Device"]


T = TypeVar("T", bound="Device")


class Device(CallbackMixIn):
    """A class that handles the TCP connection to a device.

    There are two ways to create a new device. You can create a subclass from
    `.Device` and override the `.process_message` method which handles how you
    react to a new line being received ::

        class MyDevice(Device):

            async def process_message(self, line):
                print(line)

        my_device = MyDevice('192.168.1.10', 4444)
        await my_device.start()

    Note that `.process_message` must be a coroutine. Alternatively you can
    pass a callback that will be called instead of `.process_message` when
    a new message arrives. The callback must also be a coroutine ::

        async def printer(line):
            print(line)

        my_device = MyDevice('192.168.1.10', 4444, callback=printer)
        await my_device.start()


    Parameters
    ----------
    host
        The host of the device.
    port
        The port on which the device is serving.
    callback
        The callback to call with each new message received from the client (after
        decoding into a string). If no callback is specified, `.process_message` is
        called. If the callback is not a coroutine, it will be converted to one.

    """

    def __init__(
        self,
        host: str,
        port: int,
        callback: Optional[Callable[[str], Any]] = None,
    ):
        self.host = host
        self.port = port

        # TCPStreamClient: the connection to the device.
        self._client = None
        self.listener = None

        callback = callback or self.process_message
        CallbackMixIn.__init__(self, callbacks=[callback])

    async def start(self: T) -> T:
        """Opens the connection and starts the listener."""

        if self.is_connected():
            raise RuntimeError("connection is already running.")

        self._client = await open_connection(self.host, self.port)
        self.listener = asyncio.create_task(self._listen())

        return self

    async def stop(self):
        """Closes the connection and stops the listener."""

        if self._client:
            assert self._client.writer
            self._client.writer.close()
            await self._client.writer.wait_closed()

        with contextlib.suppress(asyncio.CancelledError):
            if self.listener:
                self.listener.cancel()
                await self.listener

    def is_connected(self) -> bool:
        """Returns `True` if the connection is open."""

        if self._client is None:
            return False

        assert self._client.writer

        return not self._client.writer.is_closing()

    def write(self, message: str, newline="\n"):
        """Write to the device. The message is encoded and a new line added."""

        assert self._client and self._client.writer
        assert self.is_connected() and self._client.writer, "device is not connected"

        if not message.endswith(newline):
            message = message + newline

        self._client.writer.write(message.encode())

    async def _listen(self):
        """Listens to the reader stream and callbacks on message received."""

        if not self._client or not self._client.reader:
            raise RuntimeError("connection is not open.")

        while True:
            line = await self._client.reader.readline()
            if line == b"" or self._client.reader.at_eof():
                break
            line = line.decode().strip()
            self.notify(line)

    async def process_message(self, line: str):  # pragma: no cover
        """Processes a newly received message."""

        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-05-25
# @Filename: websocket.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import json

from typing import TYPE_CHECKING

from websockets.legacy.protocol import broadcast
from websockets.server import serve

from clu.client import AMQPClient


if TYPE_CHECKING:
    from websockets.server import WebSocketServerProtocol

    from clu.client import AMQPReply


class WebsocketServer:
    """A websocket server that allows communication with the RabbitMQ exchange.

    The websocket server is a simple pass-through between a websocket client and
    an `.AMQPClient` that connects to the RabbitMQ exchange. Any `.AMQPReply`
    received by the AMQP client is packaged as a JSON and forwarded to the
    websocket clients. Websocket clients can send messages with the format ::

        {
            "consumer": ...,
            "command_string": ...,
            "command_id": ...
        }

    that will be sent to the corresponding actor commands queue. The websocket
    server does not track command completion, which is left to the user. Including
    a ``command_id`` with the message is recommended for the client to be able to
    track commands.

    Parameters
    ----------
    whost
        The host where to run the websocket server.
    wport
        The TCP port on which to run the websocket server.
    client_kwargs
        Arguments to pass to the `.AMQPClient` connection to RabbitMQ.

    """

    def __init__(self, whost: str = "0.0.0.0", wport: int = 9876, **client_kwargs):
        self.client = AMQPClient(**client_kwargs)

        self.wparams = (whost, wport)
        self.wclients: set[WebSocketServerProtocol] = set()

    async def start(self):
        """Start the server and AMQP client."""

        self.client.add_reply_callback(self._handle_reply)
        await self.client.start()

        self.websocket_server = await serve(
            self._handle_websocket,
            *self.wparams,
        )

        return self

    async def stop(self):
        """Stop the server and AMQP client."""

        await self.client.stop()
        self.websocket_server.close()

    async def _handle_websocket(self, websocket: WebSocketServerProtocol):
        """Handle a connection to the websocket server."""

        # Register the client
        self.wclients.add(websocket)

        async for data in websocket:
            try:
                message = json.loads(data)
                if not isinstance(message, dict):
                    continue
            except ValueError:
                continue

            if "consumer" not in message or "command_string" not in message:
                continue

            command_id = message.get("command_id", None)
            await self.client.send_command(
                message["consumer"],
                message["command_string"],
                command_id=command_id,
                await_command=False,
            )

        self.wclients.remove(websocket)

    async def _handle_reply(self, reply: AMQPReply):
        """Broadcast a reply to the connected websockets."""

        message = reply.message
        data = dict(
            headers=message.headers,
            exchange=message.exchange,
            message_id=message.message_id,
            routing_key=message.routing_key,
            timestamp=message.timestamp.isoformat() if message.timestamp else None,
            body=reply.body,
        )
        broadcast(self.wclients, json.dumps(data))
